from functools import lru_cache

@lru_cache(32000)
def det_moment(kl, kr, q, x, EI, L):
    """
    See notebook in: anastruct/fem/background/primary_m_v.ipynb

    :param kl: (flt) rotational stiffness left
    :param kr: (flt) rotational stiffness right
    :param q: (flt)
    :param x: (flt) Location of bending moment
    :param EI: (flt)
    :param L: (flt) Length of the beam
    :return: (flt)
    """
    return EI*(-L**3*kl*q*(6*EI + L*kr)/(12*EI*(12*EI**2 + 4*EI*L*kl + 4*EI*L*kr + L**2*kl*kr)) +
               L*q*x*(12*EI**2 + 5*EI*L*kl + 3*EI*L*kr +
                      L**2*kl*kr)/(2*EI*(12*EI**2 + 4*EI*L*kl + 4*EI*L*kr + L**2*kl*kr)) - q*x**2/(2*EI))


@lru_cache(32000)
def det_shear(kl, kr, q, x, EI, L):
    """
    See notebook in: anastruct/fem/background/primary_m_v.ipynb

    :param kl: (flt) rotational stiffness left
    :param kr: (flt) rotational stiffness right
    :param q: (flt)
    :param x: (flt) Location of bending moment
    :param EI: (flt)
    :param L: (flt) Length of the beam
    :return: (flt)
    """
    return EI*(L*q*(12*EI**2 + 5*EI*L*kl + 3*EI*L*kr + L**2*kl*kr) /
               (2*EI*(12*EI**2 + 4*EI*L*kl + 4*EI*L*kr + L**2*kl*kr)) - q*x/EI)





from anastruct.fem.system import SystemElements
from anastruct import LoadCase, LoadCombination
import matplotlib.pyplot as plt
import numpy as np

ss = SystemElements()

height = 10

x = np.cumsum([0, 4, 7, 7, 4])
y = np.zeros(x.shape)
x = np.append(x, x[::-1])
y = np.append(y, y + height)

ss.add_element_grid(x, y)
ss.add_element([[0, 0], [0, height]])
ss.add_element([[4, 0], [4, height]])
ss.add_element([[11, 0], [11, height]])
ss.add_element([[18, 0], [18, height]])
ss.add_support_hinged([1, 5])
ss.show_structure()

lc_wind = LoadCase('wind')
lc_wind.q_load(q=-1, element_id=[10, 11, 12, 13, 5])

print(lc_wind)

ss.apply_load_case(lc_wind)
ss.show_structure()

# reset the structure
ss.remove_loads()

# create another load case
lc_cables = LoadCase('cables')
lc_cables.point_load(node_id=[2, 3, 4], Fy=-100)

combination = LoadCombination('ULS')
combination.add_load_case(lc_wind, 1.5)
combination.add_load_case(lc_cables, factor=1.2)

results = combination.solve(ss)

for k, ss in results.items():
    results[k].show_structure()
    results[k].show_displacement(show=False)
    plt.title(k)
    plt.show()



from anastruct.sectionbase.sectionbase import section_base
import math


def steel_section_properties(**kwargs):
    steel_section = kwargs.get("steelsection", 'IPE 300')
    orient = kwargs.get("orient", 'y')
    E = kwargs.get("E", 210e9)
    sw = kwargs.get("sw", False)

    param = section_base.get_section_parameters(steel_section)
    EA = E * param['Ax']
    if orient == 'y':
        EI = E * param['Iy']
    elif orient == 'z':
        EI = E * param['Iz']
    else:
        raise ValueError('Orient should be defined.')
    if sw:
        g = param['swdl']
    else:
        g = 0
    section_name = '%s(%s)' % (steel_section, orient)
    return section_name, EA, EI, g


def rectangle_properties(**kwargs):
    b = kwargs.get("b", 0.1)
    h = kwargs.get("h", 0.5)
    E = kwargs.get("E", 210e9)
    gamma = kwargs.get("gamma", 10000)
    sw = kwargs.get("sw", False)

    A = b * h
    I = b * h ** 3 / 12
    EA = E * A
    EI = E * I
    if sw:
        g = A * gamma
    else:
        g = 0
    section_name = 'rect %sx%s' % (b, h)
    return section_name, EA, EI, g


def circle_properties(**kwargs):
    d = kwargs.get("d", 0.4)
    E = kwargs.get("E", 210e9)
    gamma = kwargs.get("gamma", 10000)
    sw = kwargs.get("sw", False)

    A = math.pi * d ** 2 / 4
    I = math.pi * d ** 4 / 64
    EA = E * A
    EI = E * I
    if sw:
        g = A * gamma
    else:
        g = 0
    section_name = 'fi %s' % d
    return section_name, EA, EI, g



import math, re, collections, copy
import numpy as np
from anastruct.basic import FEMException, args_to_lists
from anastruct.fem.postprocess import SystemLevel as post_sl
from anastruct.fem.elements import Element
from anastruct.vertex import Vertex
from anastruct.fem import plotter
from . import system_components
from anastruct.vertex import vertex_range
from anastruct.sectionbase import properties


class SystemElements:
    """
    Modelling any structure starts with an object of this class.

    :ivar EA: Standard axial stiffness of elements, default=15,000
    :ivar EI: Standard bending stiffness of elements, default=5,000
    :ivar figsize: (tpl) Matplotlibs standard figure size
    :ivar element_map: (dict) Keys are the element ids, values are the element objects
    :ivar node_map: (dict) Keys are the node ids, values are the node objects.
    :ivar node_element_map: (dict) maps node ids to element objects.
    :ivar loads_point: (dict) Maps node ids to point loads.
    :ivar loads_q: (dict) Maps element ids to q-loads.
    :ivar loads_moment: (dict) Maps node ids to moment loads.
    :ivar loads_dead_load: (set) Element ids that have a dead load applied.
    """

    def __init__(self, figsize=(12, 8), EA=15e3, EI=5e3, load_factor=1, mesh=50):
        """
        * E = Young's modulus
        * A = Area
        * I = Moment of Inertia

        :param figsize: (tpl) Set the standard plotting size.
        :param EA: (flt) Standard E * A. Set the standard values of EA if none provided when generating an element.
        :param EI: (flt) Standard E * I. Set the standard values of EA if none provided when generating an element.
        :param load_factor: (flt) Multiply all loads with this factor.
        :param mesh: (int) Plotting mesh. Has no influence on the calculation.
        """
        # init object
        self.post_processor = post_sl(self)
        self.plotter = plotter.Plotter(self, mesh)
        self.plot_values = plotter.PlottingValues(self, mesh)

        # standard values if none provided
        self.EA = EA
        self.EI = EI
        self.figsize = figsize
        self.orientation_cs = -1  # needed for the loads directions

        # structure system
        self.element_map = {}  # maps element ids to the Element objects.
        self.node_map = {}  # maps node ids to the Node objects.
        self.node_element_map = {}  # maps node ids to Element objects
        # keys matrix index (for both row and columns), values K, are processed assemble_system_matrix
        self.system_spring_map = {}

        # list of indexes that remain after conditions are applied
        self._remainder_indexes = []

        # keep track of the node_id of the supports
        self.supports_fixed = []
        self.supports_hinged = []
        self.supports_roll = []
        self.supports_spring_x = []
        self.supports_spring_z = []
        self.supports_spring_y = []
        self.supports_roll_direction = []
        self.inclined_roll = {}  # map node ids to inclination angle relative to global x-axis.

        # save tuples of the arguments for copying purposes.
        self.supports_spring_args = []

        # keep track of the loads
        self.loads_point = {}  # node ids with a point loads
        self.loads_q = {}  # element ids with a q-load
        self.loads_moment = {}
        self.loads_dead_load = set()  # element ids with q-load due to dead load

        # results
        self.reaction_forces = {}  # node objects
        self.non_linear = False
        self.non_linear_elements = {}  # keys are element ids, values are dicts: {node_index: max moment capacity}
        self.buckling_factor = None

        # previous point of element
        self._previous_point = Vertex(0, 0)
        self.load_factor = load_factor

        # Objects state
        self.count = 0
        self.system_matrix_locations = []
        self.system_matrix = None
        self.system_force_vector = None
        self.system_displacement_vector = None
        self.shape_system_matrix = None
        self.reduced_force_vector = None
        self.reduced_system_matrix = None
        self._vertices = {}  # maps vertices to node ids

    @property
    def id_last_element(self):
        return max(self.element_map.keys())

    @property
    def id_last_node(self):
        return max(self.node_map.keys())

    def add_element_grid(self, x, y, EA=None, EI=None, g=None, mp=None, spring=None, **kwargs):
        """
        Add multiple elements defined by two containers with coordinates.

        :param x: (list/ np.array) x coordinates.
        :param y: (list/ np.array) y coordinates.
        :param EA: (list/ np.array) See 'add_element' method
        :param EI: (list/ np.array) See 'add_element' method
        :param g: (list/ np.array) See 'add_element' method
        :param mp: See 'add_element' method
        :param spring: See 'add_element' method
        :paramg **kwargs: See 'add_element' method
        :return: None
        """
        a = np.ones(len(x))
        if EA is None:
            EA = a * self.EA
        if EI is None:
            EI = a * self.EI
        if g is None:
            g = a * 0
        # cast to array if parameters are not given as array.
        EA = EA * a
        EI = EI * a
        g = g * a

        for i in range(len(x) - 1):
            self.add_element([[x[i], y[i]], [x[i + 1], y[i + 1]]], EA[i], EI[i], g[i], mp, spring, **kwargs)

    def add_truss_element(self, location, EA=None, **kwargs):
        """
        .. highlight:: python

        Add an element that only has axial force.

        :param location: (list/ Vertex) The two nodes of the element or the next node of the element.

            :Example:

            .. code-block:: python

                   location=[[x, y], [x, y]]
                   location=[Vertex, Vertex]
                   location=[x, y]
                   location=Vertex

        :param EA: (flt) EA
        :return: (int) Elements ID.
        """
        return self.add_element(location, EA, element_type='truss',  **kwargs)

    def add_element(self, location, EA=None, EI=None, g=0, mp=None, spring=None, **kwargs):
        """
        :param location: (list/ Vertex) The two nodes of the element or the next node of the element.

            :Example:

            .. code-block:: python

                   location=[[x, y], [x, y]]
                   location=[Vertex, Vertex]
                   location=[x, y]
                   location=Vertex

        :param EA: (flt) EA
        :param EI: (flt) EI
        :param g: (flt) Weight per meter. [kN/m] / [N/m]
        :param mp: (dict) Set a maximum plastic moment capacity. Keys are integers representing the nodes. Values
                          are the bending moment capacity.

            :Example:

            .. code-block:: python

                mp={1: 210e3,
                    2: 180e3}

        :param spring: (dict) Set a rotational spring or a hinge (k=0) at node 1 or node 2.

            :Example:

            .. code-block:: python

                spring={1: k
                        2: k}


                # Set a hinged node:
                spring={1: 0}


        :return: (int) Elements ID.
        """
        element_type = kwargs.get("element_type", "general")

        EA = self.EA if EA is None else EA
        EI = self.EI if EI is None else EI
        
        section_name = ''
        # change EA EI and g if steel section specified
        if 'steelsection' in kwargs:
            section_name, EA, EI, g = properties.steel_section_properties(**kwargs)
        # change EA EI and g if rectangle section specified
        if 'h' in kwargs:
            section_name, EA, EI, g = properties.rectangle_properties(**kwargs)
        # change EA EI and g if circle section specified
        if 'd' in kwargs:
            section_name, EA, EI, g = properties.circle_properties(**kwargs)
        
        if element_type == 'truss':
            EI = 1e-14

        # add the element number
        self.count += 1

        point_1, point_2 = system_components.util.det_vertices(self, location)
        node_id1, node_id2 = system_components.util.det_node_ids(self, point_1, point_2)

        point_1, point_2, node_id1, node_id2, spring, mp, angle = \
            system_components.util.force_elements_orientation(point_1, point_2, node_id1, node_id2, spring, mp)

        system_components.util.append_node_id(self, point_1, point_2, node_id1, node_id2)
        system_components.util.ensure_single_hinge(self, spring, node_id1, node_id2)

        # add element
        element = Element(self.count, EA, EI, (point_2 - point_1).modulus(), angle, point_1, point_2, spring)
        element.node_id1 = node_id1
        element.node_id2 = node_id2
        element.node_map = {node_id1: self.node_map[node_id1],
                            node_id2: self.node_map[node_id2]}
        
        # needed for element annotations
        element.section_name = section_name

        element.type = element_type

        self.element_map[self.count] = element

        for node in (node_id1, node_id2):
            if node in self.node_element_map:
                self.node_element_map[node].append(element)
            else:
                self.node_element_map[node] = [element]

        # Register the elements per node
        for node_id in (node_id1, node_id2):
            self.node_map[node_id].elements[element.id] = element

        if mp is not None:
            assert type(mp) == dict, "The mp parameter should be a dictionary."
            self.non_linear_elements[element.id] = mp
            self.non_linear = True
        system_components.assembly.dead_load(self, g, element.id)

        return self.count

    def add_multiple_elements(self, location, n=None, dl=None, EA=None, EI=None, g=0, mp=None, spring=None,
                              **kwargs):
        """
        Add multiple elements defined by the first and the last point.

        :param location: See 'add_element' method
        :param n: (int) Number of elements.
        :param dl: (flt) Distance between the elements nodes.
        :param EA: See 'add_element' method
        :param EI: See 'add_element' method
        :param g: See 'add_element' method
        :param mp: See 'add_element' method
        :param spring: See 'add_element' method

        **Keyword Args:**

        :param element_type: (str) See 'add_element' method
        :param first: (dict) Different arguments for the first element
        :param last: (dict) Different arguments for the last element
        :param steelsection: (str) Steel section name like IPE 300
        :param orient: (str) Steel section axis for moment of inertia - 'y' and 'z' possible
        :param b: (flt) Width of generic rectangle section
        :param h: (flt) Height of generic rectangle section
        :param d: (flt) Diameter of generic circle section
        :param sw: (boolean) If true self weight of section is considered as dead load
        :param E: (flt) Modulus of elasticity for section material
        :param gamma: (flt) Weight of section material per volume unit. [kN/m3] / [N/m3]s

            :Example:

            .. code-block:: python

                last={'EA': 1e3, 'mp': 290}

        :return: (list) Element IDs
        """

        first = kwargs.get("first", {})
        last = kwargs.get("last", {})
        element_type = kwargs.get("element_type", "general")

        for el in (first, last):
            if "EA" not in el:
                el["EA"] = EA
            if "EI" not in el:
                el["EI"] = EI
            if "g" not in el:
                el["g"] = g
            if "mp" not in el:
                el["mp"] = mp
            if "spring" not in el:
                el["spring"] = spring
            if "element_type" not in el:
                el["element_type"] = element_type

        point_1, point_2 = system_components.util.det_vertices(self, location)
        length = (point_2 - point_1).modulus()
        direction = (point_2 - point_1).unit()

        if type(n) == type(dl):
            raise FEMException("Wrong parameters", "One, and only one, of n and dl should be passed as argument.")
        elif n:
            dl = length / n

        point = point_1 + direction * dl
        elements = [
            self.add_element((point_1, point), first["EA"], first["EI"], first["g"], first["mp"], first["spring"],
                             element_type=first["element_type"], **kwargs)]

        l = 2 * dl
        while l < length:
            point += direction * dl
            elements.append(self.add_element(point, EA, EI, g, mp, spring, element_type=element_type, **kwargs))
            l += dl

        elements.append(self.add_element(point_2, last["EA"], last["EI"], last["g"], last["mp"], last["spring"],
                                         element_type=last["element_type"], **kwargs))
        return elements

    def insert_node(self, element_id, location=None, factor=None):
        """
        Insert a node into an existing structure.
        This can be done by adding a new Vertex at any given location, or by setting a factor of the elements
        length. E.g. if you want a node at 40% of the elements length, you pass factor = 0.4.

        Note: this method completely rebuilds the SystemElements object and is therefore slower then building
        a model with `add_element` methods.

        :param element_id: (int) Id number of the element you want to insert the node.
        :param location: (list/ Vertex) The nodes of the element or the next node of the element.

            :Example:

            .. code-block:: python

                   location=[x, y]
                   location=Vertex

        :param: factor: (flt) Value between 0 and 1 to determine the new node location.
        """
        ss = SystemElements(EA=self.EA, EI=self.EI, load_factor=self.load_factor,
                            mesh=self.plotter.mesh)

        for element in self.element_map.values():
            g = self.element_map[element.id].dead_load
            mp = self.non_linear_elements[element.id] if element.id in self.non_linear_elements else None
            if element_id == element.id:
                if factor is not None:
                    location = factor * (element.vertex_2 - element.vertex_1) + element.vertex_1
                else:
                    location = Vertex(location)

                mp1 = mp2 = spring1 = spring2 = None
                if mp is not None:
                    if 1 in mp:
                        mp1 = {1: mp[1]}
                    if 2 in mp:
                        mp2 = {2: mp[2]}
                if element.springs is not None:
                    if 1 in element.springs:
                        spring1 = {1: element.springs[1]}
                    if 2 in element.springs:
                        spring2 = {2: element.springs[2]}

                ss.add_element([element.vertex_1, location], EA=element.EA, EI=element.EI, g=g, mp=mp1, spring=spring1)
                ss.add_element([location, element.vertex_2], EA=element.EA, EI=element.EI, g=g, mp=mp2, spring=spring2)

            else:
                ss.add_element([element.vertex_1, element.vertex_2], EA=element.EA,
                               EI=element.EI, g=g, mp=mp, spring=element.springs)
        self.__dict__ = ss.__dict__.copy()

    def solve(self, force_linear=False, verbosity=0, max_iter=200, geometrical_non_linear=False, **kwargs):

        """
        Compute the results of current model.

        :param force_linear: (bool) Force a linear calculation. Even when the system has non linear nodes.
        :param verbosity: (int) 0. Log calculation outputs. 1. silence.
        :param max_iter: (int) Maximum allowed iterations.
        :param geometrical_non_linear: (bool) Calculate second order effects and determine the buckling factor.
        :return: (array) Displacements vector.


        Development **kwargs:
            :param naked: (bool) Whether or not to run the solve function without doing post processing.
            :param discretize_kwargs: When doing a geometric non linear analysis you can reduce or increase the number
                                      of elements created that are used for determining the buckling_factor
        """

        # kwargs: arguments for the iterative solver callers such as the _stiffness_adaptation method.
        #                naked (bool) Default = False, if True force lines won't be computed.

        if self.system_displacement_vector is None:
            system_components.assembly.process_supports(self)

        naked = kwargs.get("naked", False)

        if not naked:
            if not self.validate():
                if all(['general' in element.type for element in self.element_map.values()]):
                    raise FEMException('StabilityError', 'The eigenvalues of the stiffness matrix are non zero, '
                                                         'which indicates a instable structure. '
                                                         'Check your support conditions')

        # (Re)set force vectors
        for el in self.element_map.values():
            el.reset()
        system_components.assembly.prep_matrix_forces(self)
        assert (self.system_force_vector is not None), "There are no forces on the structure"

        if self.non_linear and not force_linear:
            return system_components.solver.stiffness_adaptation(self, verbosity, max_iter)

        system_components.assembly.assemble_system_matrix(self)
        if geometrical_non_linear:
            discretize_kwargs = kwargs.get('discretize_kwargs', None)
            self.buckling_factor = system_components.solver.geometrically_non_linear(self, verbosity,
                                                                                     discretize_kwargs=discretize_kwargs)
            return self.system_displacement_vector

        system_components.assembly.process_conditions(self)

        # solution of the reduced system (reduced due to support conditions)
        reduced_displacement_vector = np.linalg.solve(self.reduced_system_matrix, self.reduced_force_vector)

        # add the solution of the reduced system in the complete system displacement vector
        self.system_displacement_vector = np.zeros(self.shape_system_matrix)
        np.put(self.system_displacement_vector, self._remainder_indexes, reduced_displacement_vector)

        # determine the displacement vector of the elements
        for el in self.element_map.values():
            index_node_1 = (el.node_1.id - 1) * 3
            index_node_2 = (el.node_2.id - 1) * 3

            # node 1 ux, uz, phi
            el.element_displacement_vector[:3] = self.system_displacement_vector[index_node_1: index_node_1 + 3]
            # node 2 ux, uz, phi
            el.element_displacement_vector[3:] = self.system_displacement_vector[index_node_2: index_node_2 + 3]
            el.determine_force_vector()

        if not naked:
            # determining the node results in post processing class
            self.post_processor.node_results_elements()
            self.post_processor.node_results_system()
            self.post_processor.reaction_forces()
            self.post_processor.element_results()

            # check the values in the displacement vector for extreme values, indicating a flawed calculation
            assert (np.any(self.system_displacement_vector < 1e6)), "The displacements of the structure exceed 1e6. " \
                                                                    "Check your support conditions," \
                                                                    "or your elements Young's modulus"

        return self.system_displacement_vector

    def validate(self, min_eigen=1e-9):
        """
        Validate the stability of the stiffness matrix.

        :param min_eigen: (flt) Minimum value of the eigenvalues of the stiffness matrix. This value should be close
        to zero.
        :return: (bool)
        """

        ss = copy.copy(self)
        system_components.assembly.prep_matrix_forces(ss)
        assert (np.abs(ss.system_force_vector).sum() != 0), "There are no forces on the structure"
        ss._remainder_indexes = []
        system_components.assembly.assemble_system_matrix(ss)

        system_components.assembly.process_conditions(ss)

        w, _ = np.linalg.eig(ss.reduced_system_matrix)
        return np.all(w > min_eigen)

    def add_support_hinged(self, node_id):
        """
        Model a hinged support at a given node.

        :param node_id: (int/ list) Represents the nodes ID
        """
        if not isinstance(node_id, collections.Iterable):
            node_id = (node_id,)

        for id_ in node_id:
            id_ = _negative_index_to_id(id_, self.node_map.keys())
            system_components.util.support_check(self, id_)

            # add the support to the support list for the plotter
            self.supports_hinged.append(self.node_map[id_])

    def add_support_roll(self, node_id, direction='x', angle=None):
        """
        Adds a rolling support at a given node.

        :param node_id: (int/ list) Represents the nodes ID
        :param direction: (str/ list) Represents the direction that is free: 'x', 'y'
        :param angle (flt/ list) Angle in degrees relative to global x-axis.
                                If angle is given, the support will be inclined.
        """
        if not isinstance(node_id, collections.Iterable):
            node_id = (node_id,)

        for id_ in node_id:
            id_ = _negative_index_to_id(id_, self.node_map.keys())
            system_components.util.support_check(self, id_)

            if direction == 'x':
                direction = 2
            elif direction == 'y':
                direction = 1

            if angle is not None:
                direction = 2
                self.inclined_roll[id_] = np.radians(-angle)

            # add the support to the support list for the plotter
            self.supports_roll.append(self.node_map[id_])
            self.supports_roll_direction.append(direction)

    def add_support_fixed(self, node_id):
        """
        Add a fixed support at a given node.

        :param node_id: (int/ list) Represents the nodes ID
        """
        if not isinstance(node_id, collections.Iterable):
            node_id = (node_id,)

        for id_ in node_id:
            id_ = _negative_index_to_id(id_, self.node_map.keys())
            system_components.util.support_check(self, id_)

            # add the support to the support list for the plotter
            self.supports_fixed.append(self.node_map[id_])

    def add_support_spring(self, node_id, translation, k, roll=False):
        """
        Add a translational support at a given node.

        :param translation: (int/ list) Represents the prevented translation.

            **Note**

            |  1 = translation in x
            |  2 = translation in z
            |  3 = rotation in y

        :param node_id: (int/ list) Integer representing the nodes ID.
        :param k: (flt) Stiffness of the spring
        :param roll: (bool) If set to True, only the translation of the spring is controlled.

        """
        self.supports_spring_args.append((node_id, translation, k, roll))
        # The stiffness of the spring is added in the system matrix at the location that represents the node and the
        # displacement.
        if not isinstance(node_id, collections.Iterable):
            node_id = (node_id,)

        for id_ in node_id:
            id_ = _negative_index_to_id(id_, self.node_map.keys())
            system_components.util.support_check(self, id_)

            # determine the location in the system matrix
            # row and column are the same
            matrix_index = (id_ - 1) * 3 + translation - 1

            self.system_spring_map[matrix_index] = k

            # add the support to the support list for the plotter
            if translation == 1:
                self.supports_spring_x.append((self.node_map[id_], roll))
            elif translation == 2:
                self.supports_spring_z.append((self.node_map[id_], roll))
            else:
                self.supports_spring_y.append((self.node_map[id_], roll))

    def q_load(self, q, element_id, direction="element"):
        """
        Apply a q-load to an element.

        :param element_id: (int/ list) representing the element ID
        :param q: (flt) value of the q-load
        :param direction: (str) "element", "x", "y"
        """
        q, element_id, direction = args_to_lists(q, element_id, direction)

        for i in range(len(element_id)):
            id_ = _negative_index_to_id(element_id[i], self.element_map.keys())
            self.plotter.max_q = max(self.plotter.max_q, abs(q[i]))
            self.loads_q[id_] = q[i] * self.orientation_cs * self.load_factor
            el = self.element_map[id_]
            el.q_load = q[i] * self.orientation_cs * self.load_factor
            el.q_direction = direction[i]

    def point_load(self, node_id, Fx=0, Fy=0, rotation=0):
        """
        Apply a point load to a node.

        :param node_id: (int/ list) Nodes ID.
        :param Fx: (flt/ list) Force in global x direction.
        :param Fy: (flt/ list) Force in global x direction.
        :param rotation: (flt/ list) Rotate the force clockwise. Rotation is in degrees.
        """
        node_id, Fx, Fy, rotation = args_to_lists(node_id, Fx, Fy, rotation)

        for i in range(len(node_id)):
            id_ = _negative_index_to_id(node_id[i], self.node_map.keys())
            self.plotter.max_system_point_load = max(self.plotter.max_system_point_load,
                                                     (Fx[i] ** 2 + Fy[i] ** 2) ** 0.5)
            cos = math.cos(math.radians(rotation[i]))
            sin = math.sin(math.radians(rotation[i]))
            self.loads_point[id_] = (Fx[i] * cos + Fy[i] * sin, Fy[i] * self.orientation_cs * cos + Fx[i] * sin)

    def moment_load(self, node_id, Ty):
        """
        Apply a moment on a node.

        :param node_id: (int/ list) Nodes ID.
        :param Ty: (flt/ list) Moments acting on the node.
        """
        node_id, Ty = args_to_lists(node_id, Ty)

        for i in range(len(node_id)):
            id_ = _negative_index_to_id(node_id[i], self.node_map.keys())
            self.loads_moment[id_] = Ty[i]

    def show_structure(self, verbosity=0, scale=1., offset=(0, 0), figsize=None, show=True, supports=True,
                       values_only=False, annotations=False):
        """
        Plot the structure.

        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :param supports: (bool) Show the supports.
        :param values_only: (bool) Return the values that would be plotted as tuple containing two arrays: (x, y)
        :param annotations: (boolean) if True, structure annotations are plotted. It includes section name.
                                      Note: only works when verbosity is equal to 0.
        :return: (figure)
        """
        figsize = self.figsize if figsize is None else figsize
        if values_only:
            return self.plot_values.structure()
        return self.plotter.plot_structure(figsize, verbosity, show, supports, scale, offset, annotations=annotations)

    def show_bending_moment(self, factor=None, verbosity=0, scale=1, offset=(0, 0), figsize=None, show=True,
                            values_only=False):
        """
        Plot the bending moment.

        :param factor: (flt) Influence the plotting scale.
        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :param values_only: (bool) Return the values that would be plotted as tuple containing two arrays: (x, y)
        :return: (figure)
        """
        if values_only:
            return self.plot_values.bending_moment(factor)
        figsize = self.figsize if figsize is None else figsize
        return self.plotter.bending_moment(factor, figsize, verbosity, scale, offset, show)

    def show_axial_force(self, factor=None, verbosity=0, scale=1, offset=(0, 0), figsize=None, show=True,
                         values_only=False):
        """
        Plot the axial force.

        :param factor: (flt) Influence the plotting scale.
        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :param values_only: (bool) Return the values that would be plotted as tuple containing two arrays: (x, y)
        :return: (figure)
        """
        if values_only:
            return self.plot_values.axial_force(factor)
        figsize = self.figsize if figsize is None else figsize
        return self.plotter.axial_force(factor, figsize, verbosity, scale, offset, show)

    def show_shear_force(self, factor=None, verbosity=0, scale=1, offset=(0, 0), figsize=None, show=True,
                         values_only=False):
        """
        Plot the shear force.
        :param factor: (flt) Influence the plotting scale.
        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :param values_only: (bool) Return the values that would be plotted as tuple containing two arrays: (x, y)
        :return: (figure)
        """
        if values_only:
            return self.plot_values.shear_force(factor)
        figsize = self.figsize if figsize is None else figsize
        return self.plotter.shear_force(factor, figsize, verbosity, scale, offset, show)

    def show_reaction_force(self, verbosity=0, scale=1, offset=(0, 0), figsize=None, show=True):
        """
        Plot the reaction force.

        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :return: (figure)
        """
        figsize = self.figsize if figsize is None else figsize
        return self.plotter.reaction_force(figsize, verbosity, scale, offset, show)

    def show_displacement(self, factor=None, verbosity=0, scale=1, offset=(0, 0), figsize=None, show=True,
                          linear=False, values_only=False):
        """
        Plot the displacement.

        :param factor: (flt) Influence the plotting scale.
        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :param linear: (bool) Don't evaluate the displacement values in between the elements
        :param values_only: (bool) Return the values that would be plotted as tuple containing two arrays: (x, y)
        :return: (figure)
        """
        if values_only:
            return self.plot_values.displacements(factor, linear)
        figsize = self.figsize if figsize is None else figsize
        return self.plotter.displacements(factor, figsize, verbosity, scale, offset, show, linear)

    def show_results(self, verbosity=0, scale=1, offset=(0, 0), figsize=None, show=True):
        """
        Plot all the results in one window.

        :param verbosity: (int) 0: All information, 1: Suppress information.
        :param scale: (flt) Scale of the plot.
        :param offset: (tpl) Offset the plots location on the figure.
        :param figsize: (tpl) Change the figure size.
        :param show: (bool) Plot the result or return a figure.
        :return: (figure)
        """
        figsize = self.figsize if figsize is None else figsize
        return self.plotter.results_plot(figsize, verbosity, scale, offset, show)

    def get_node_results_system(self, node_id=0):
        """
        These are the node results. These are the opposite of the forces and displacements working on the elements and
        may seem counter intuitive.

        :param node_id: (integer) representing the node's ID. If integer = 0, the results of all nodes are returned
        :return:

        |  if node_id == 0: (list)
        |
        |    Returns a list containing tuples with the results:

        ::

            [(id, Fx, Fy, Ty, ux, uy, phi_y), (id, Fx, Fy...), () .. ]
        |
        |  if node_id > 0: (dict)
        """
        result_list = []
        if node_id != 0:
            node_id = _negative_index_to_id(node_id, self.node_map)
            node = self.node_map[node_id]
            return {
                "id": node.id,
                "Fx": node.Fx,
                "Fy": node.Fy,
                "Ty": node.Ty,
                "ux": node.ux,
                "uy": -node.uz,
                "phi_y": node.phi_y
            }
        else:
            for node in self.node_map.values():
                result_list.append((node.id, node.Fx, node.Fy, node.Ty, node.ux, -node.uz, node.phi_y))
        return result_list

    def get_node_displacements(self, node_id=0):
        """
        :param node_id: (int) Represents the node's ID. If integer = 0, the results of all nodes are returned.
        :return:
        |  if node_id == 0: (list)
        |
        |    Returns a list containing tuples with the results:

        ::

             [(id, ux, uy, phi_y), (id, ux, uy, phi_y),  ... (id, ux, uy, phi_y) ]
        |
        |  if node_id > 0: (dict)
        """
        result_list = []
        if node_id != 0:
            node = self.node_map[node_id]
            return {
                "id": node.id,
                "ux": -node.ux,
                "uy": node.uz,  # - * -  = +
                "phi_y": node.phi_y
            }
        else:
            for node in self.node_map.values():
                result_list.append((node.id, -node.ux, node.uz, node.phi_y))
        return result_list

    def get_element_results(self, element_id=0, verbose=False):
        """
        :param element_id: (int) representing the elements ID. If elementID = 0 the results of all elements are returned.
        :param verbose: (bool) If set to True the numerical results for the deflection and the bending moments are
                               returned.

        :return:
        |
        |  if node_id == 0: (list)
        |
        |    Returns a list containing tuples with the results:

        ::

             [(id, length, alpha, u, N_1, N_2), (id, length, alpha, u, N_1, N_2), ... (id, length, alpha, u, N_1, N_2)]
        |
        |  if node_id > 0: (dict)
        |

        """
        if element_id != 0:
            element_id = _negative_index_to_id(element_id, self.element_map)
            el = self.element_map[element_id]
            if el.type == "truss":
                return {
                    "id": el.id,
                    "length": el.l,
                    "alpha": el.angle,
                    "u": el.extension[0],
                    "N": el.N_1,
                }
            else:
                return {
                    "id": el.id,
                    "length": el.l,
                    "alpha": el.angle,
                    "u": el.extension[0],
                    "N": el.N_1,
                    "wmax": np.min(el.deflection),
                    "wmin": np.max(el.deflection),
                    "w": el.deflection if verbose else None,
                    "Mmin": np.min(el.bending_moment),
                    "Mmax": np.max(el.bending_moment),
                    "M": el.bending_moment if verbose else None,
                    "q": el.q_load
                }
        else:
            result_list = []
            for el in self.element_map.values():

                if el.type == "truss":
                    result_list.append({
                        "id": el.id,
                        "length": el.l,
                        "alpha": el.angle,
                        "u": el.extension[0],
                        "N": el.N_1,
                    }
                    )

                else:
                    result_list.append(
                        {
                            "id": el.id,
                            "length": el.l,
                            "alpha": el.angle,
                            "u": el.extension[0],
                            "N": el.N_1,
                            "wmax": np.min(el.deflection),
                            "wmin": np.max(el.deflection),
                            "w": el.deflection if verbose else None,
                            "Mmin": np.min(el.bending_moment),
                            "Mmax": np.max(el.bending_moment),
                            "M": el.bending_moment if verbose else None,
                            "q": el.q_load
                        }
                    )
            return result_list

    def get_element_result_range(self, unit):
        """
        Useful when added lots of elements. Returns a list of all the queried unit.

        :param unit: (str)
            - 'shear'
            - 'moment'
            - 'axial'

        :return: (list)
        """
        if unit == "shear":
            return [el.shear_force[0] for el in self.element_map.values()]
        elif unit == "moment":
            return [el.bending_moment[0] for el in self.element_map.values()]
        elif unit == "axial":
            return [el.N_1 for el in self.element_map.values()]

    def get_node_result_range(self, unit):
        """
        Query a list with node results.

           :param unit: (str)
            - 'uy'
            - 'ux'
            - 'phi_y'
        :return: (list)
        """
        if unit == "uy":
            return [node.uz for node in self.node_map.values()]  # - * -  = +
        elif unit == "ux":
            return [-node.ux for node in self.node_map.values()]
        elif unit == "phi_y":
            return [node.phi_y for node in self.node_map.values()]

    def find_node_id(self, vertex):
        """
        Retrieve the ID of a certain location.

        :param vertex: (Vertex/ list/ tpl) Vertex_xz, [x, y], (x, y)
        :return: (int/ None) id of the node at the location of the vertex
        """
        if isinstance(vertex, (list, tuple)):
            vertex = Vertex(vertex)
        try:
            tol = 1e-9
            return next(filter(lambda x: math.isclose(x.vertex.x, vertex.x, abs_tol=tol)
                                         and math.isclose(x.vertex.y, vertex.y, abs_tol=tol),
                               self.node_map.values())).id
        except StopIteration:
            return None

    def nodes_range(self, dimension):
        """
        Retrieve a list with coordinates x or z (y).

        :param dimension: (str) "both", 'x', 'y' or 'z'
        :return: (list)
        """
        return list(
            map(
                lambda x: x.vertex.x if dimension == 'x'
                else x.vertex.z if dimension == 'z' else x.vertex.y if dimension == 'y'
                else (x.vertex.x, x.vertex.y) if dimension == "both"
                else None,
                self.node_map.values()))

    def nearest_node(self, dimension, val):
        """
        Retrieve the nearest node ID.

        :param dimension: (str) "both", 'x', 'y' or 'z'
        :param val: (flt) Value of the dimension.
        :return: (int) ID of the node.
        """
        if dimension == "both":
            match = list(map(lambda x: x[1],
                             filter(lambda x: x[0][0] == val[0] and x[0][1] == val[1],
                                    zip(self.nodes_range("both"), self.node_map.keys())
                                    )
                             )
                         )
            return match[0] if len(match) > 0 else None
        else:
            return np.argmin(np.abs(np.array(self.nodes_range(dimension)) - val))

    def discretize(self, n=10):
        """
        Takes an already defined :class:`.SystemElements` object and increases the number of elements.

        :param n: (int) Divide the elements into n sub-elements.
        """
        ss = SystemElements(EA=self.EA, EI=self.EI, load_factor=self.load_factor,
                            mesh=self.plotter.mesh)

        for element in self.element_map.values():
            g = self.element_map[element.id].dead_load
            mp = self.non_linear_elements[element.id] if element.id in self.non_linear_elements else None

            for i, v in enumerate(vertex_range(element.vertex_1, element.vertex_2, n)):
                if i == 0:
                    last_v = v
                    continue

                loc = [last_v, v]
                ss.add_element(loc, EA=element.EA, EI=element.EI, g=g, mp=mp, spring=element.springs)
                last_v = v

        # supports
        for node, direction in zip(self.supports_roll, self.supports_roll_direction):
            ss.add_support_roll((node.id - 1) * n + 1, direction)
        for node in self.supports_fixed:
            ss.add_support_fixed((node.id - 1) * n + 1)
        for node in self.supports_hinged:
            ss.add_support_hinged((node.id - 1) * n + 1)
        for args in self.supports_spring_args:
            ss.add_support_spring((args[0] - 1) * n + 1, *args[1:])

        # loads
        for node_id, forces in self.loads_point.items():
            ss.point_load((node_id - 1) * n + 1, Fx=forces[0], Fy=forces[1] / self.orientation_cs)
        for node_id, forces in self.loads_moment.items():
            ss.moment_load((node_id - 1) * n + 1, forces)
        for element_id, forces in self.loads_q.items():
            ss.q_load(q=forces / self.orientation_cs / self.load_factor,
                      element_id=element_id,
                      direction=self.element_map[element_id].q_direction)

        self.__dict__ = ss.__dict__.copy()

    def remove_loads(self, dead_load=False):
        """
        Remove all the applied loads from the structure.

        :param dead_load: (bool) Remove the dead load.
        """

        self.loads_point = {}
        self.loads_q = {}
        self.loads_moment = {}

        for k in self.element_map:
            self.element_map[k].q_load = 0
            if dead_load:
                self.element_map[k].dead_load = 0
        if dead_load:
            self.loads_dead_load = set()

    def apply_load_case(self, loadcase):
        """
        :param loadcase:
        :return:
        """
        for method, kwargs in loadcase.spec.items():
            method = method.split('-')[0]
            kwargs = re.sub(r"[{}]", '', str(kwargs))
            # pass the groups that match back to the replace
            kwargs = re.sub(r".??(\w+).?:", r'\1=', kwargs)

            exec('self.{}({})'.format(method, kwargs))

    def __deepcopy__(self, memo):
        system = copy.copy(self)
        mesh = self.plotter.mesh
        system.plotter = None
        system.post_processor = None
        system.plot_values = None

        system.__dict__ = copy.deepcopy(system.__dict__)
        system.plotter = plotter.Plotter(system, mesh)
        system.post_processor = post_sl(system)
        system.plot_values = plotter.PlottingValues

        return system


def _negative_index_to_id(idx, collection):
    if idx > 0:
        return idx
    else:
        return max(collection) + (idx + 1)





from anastruct.fem.system import SystemElements


ss = SystemElements()
ss.add_element(location=[[0, 0], [3, 4]], EA=5e9, EI=8000)
ss.add_element(location=[[3, 4], [8, 4]], EA=5e9, EI=4000)

ss.q_load(element_id=2, q=-10)

ss.add_support_hinged(node_id=1)
ss.add_support_fixed(node_id=3)

ss.solve()
ss.show_structure()
ss.show_reaction_force()
ss.show_axial_force()
ss.show_shear_force()
ss.show_bending_moment()
ss.show_displacement()





from anastruct.fem.system import SystemElements

ss = SystemElements(EI=5e3, EA=1e5)

ss.add_element([0, 10])

ss.add_support_hinged(1)
ss.add_support_roll(2, 1)

ss.point_load(2, Fy=-1)

if __name__ == "__main__":
    ss.solve(geometrical_non_linear=True, discretize_kwargs=dict(n=15))
    print(ss.buckling_factor)
    ss.show_displacement()



from anastruct.fem.elements import det_moment, det_shear
import numpy as np
import math


def set_force_vector(system, force_list):
    """
    :param force_list: list containing tuples with the
    1. number of the node,
    2. the number of the direction (1 = x, 2 = z, 3 = y)
    3. the force
    [(1, 3, 1000)] node=1, direction=3 (y), force=1000
    list may contain multiple tuples
    :return: Vector with forces on the nodes
    """
    for id_, direction, force in force_list:
        system.system_force_vector[(id_ - 1) * 3 + direction - 1] += force

    return system.system_force_vector


def prep_matrix_forces(system):
    system.system_force_vector = system.system_force_vector = np.zeros(len(system._vertices) * 3)
    apply_perpendicular_q_load(system)
    apply_point_load(system)
    apply_moment_load(system)


def apply_moment_load(system):
    for node_id, Ty in system.loads_moment.items():
        set_force_vector(system, [(node_id, 3, Ty * system.load_factor)])


def apply_point_load(system):
    for node_id in system.loads_point:
        Fx, Fz = system.loads_point[node_id]
        # system force vector.
        set_force_vector(system, [(node_id, 1, Fx * system.load_factor), (node_id, 2, Fz * system.load_factor)])


def apply_perpendicular_q_load(system):
    for element_id in system.loads_dead_load:
        element = system.element_map[element_id]
        q_perpendicular = element.all_q_load

        if q_perpendicular == 0:
            continue
        elif element.q_direction == "x" or element.q_direction == "y" or element.dead_load:
            apply_parallel_q_load(system, element)

        kl = element.constitutive_matrix[1][1] * 1e6
        kr = element.constitutive_matrix[2][2] * 1e6

        if math.isclose(kl, kr):
            left_moment = det_moment(kl, kr, q_perpendicular, 0, element.EI, element.l)
            right_moment = -left_moment
            rleft = det_shear(kl, kr, q_perpendicular, 0, element.EI, element.l)
            rright = rleft
        else:
            # minus because of systems positive rotation
            left_moment = det_moment(kl, kr, q_perpendicular, 0, element.EI, element.l)
            right_moment = -det_moment(kl, kr, q_perpendicular, element.l, element.EI, element.l)
            rleft = det_shear(kl, kr, q_perpendicular, 0, element.EI, element.l)
            rright = -det_shear(kl, kr, q_perpendicular, element.l, element.EI, element.l)

        rleft_x = rleft * math.sin(element.angle)
        rright_x = rright * math.sin(element.angle)

        rleft_z = rleft * math.cos(element.angle)
        rright_z = rright * math.cos(element.angle)

        if element.type == 'truss':
            left_moment = 0
            right_moment = 0

        primary_force = np.array([rleft_x, rleft_z, left_moment, rright_x, rright_z, right_moment])
        element.element_primary_force_vector -= primary_force

        # Set force vector
        system.system_force_vector[(element.node_id1 - 1) * 3: (element.node_id1 - 1) * 3 + 3] += primary_force[0:3]
        system.system_force_vector[(element.node_id2 - 1) * 3: (element.node_id2 - 1) * 3 + 3] += primary_force[3:]


def apply_parallel_q_load(system, element):
    direction = element.q_direction
    # dead load
    factor_dl = abs(math.sin(element.angle))

    def update(Fx, Fz):
        element.element_primary_force_vector[0] -= Fx
        element.element_primary_force_vector[1] -= Fz
        element.element_primary_force_vector[3] -= Fx
        element.element_primary_force_vector[4] -= Fz

        set_force_vector(system, [
            (element.node_1.id, 2, Fz), (element.node_2.id, 2, Fz),
            (element.node_1.id, 1, Fx), (element.node_2.id, 1, Fx)])

    if direction == "x":
        factor = abs(math.cos(element.angle))

        for q_element in (element.q_load * factor, element.dead_load * factor_dl):
            # q_load working at parallel to the elements x-axis          # set the proper direction
            Fx = -q_element * math.cos(element.angle) * element.l * 0.5
            Fz = q_element * abs(math.sin(element.angle)) * element.l * 0.5 * np.sign(math.sin(element.angle))

            update(Fx, Fz)

    else:
        if math.isclose(element.angle, 0):
            # horizontal element cannot have parallel forces due to self weight or q-load in y direction.
            return None
        factor = abs(math.sin(element.angle))

        for q_element in (element.q_load * factor, element.dead_load * factor_dl):
            # q_load working at parallel to the elements x-axis          # set the proper direction
            Fx = q_element * math.cos(element.angle) * element.l * 0.5 * -np.sign(math.sin(element.angle))
            Fz = q_element * abs(math.sin(element.angle)) * element.l * 0.5

            update(Fx, Fz)


def dead_load(system, g, element_id):
    system.loads_dead_load.add(element_id)
    system.element_map[element_id].dead_load = g


def assemble_system_matrix(system, validate=False, geometric_matrix=False):
    """
    Shape of the matrix = n nodes * n d.o.f.
    Shape = n * 3
    """
    system._remainder_indexes = []
    if not geometric_matrix:
        shape = len(system.node_map) * 3
        system.shape_system_matrix = shape
        system.system_matrix = np.zeros((shape, shape))

    for matrix_index, K in system.system_spring_map.items():
        #  first index is row, second is column
        system.system_matrix[matrix_index][matrix_index] += K

    # Determine the elements location in the stiffness matrix.
    # system matrix [K]
    #
    # [fx 1] [K         |  \ node 1 starts at row 1
    # |fz 1] |  K       |  /
    # |Ty 1] |   K      | /
    # |fx 2] |    K     |  \ node 2 starts at row 4
    # |fz 2] |     K    |  /
    # |Ty 2] |      K   | /
    # |fx 3] |       K  |  \ node 3 starts at row 7
    # |fz 3] |        K |  /
    # [Ty 3] [         K] /
    #
    #         n   n  n
    #         o   o  o
    #         d   d  d
    #         e   e  e
    #         1   2  3
    #
    # thus with appending numbers in the system matrix: column = row

    for i in range(len(system.element_map)):
        element = system.element_map[i + 1]
        element_matrix = element.stiffness_matrix

        # n1 and n2 are starting indexes of the rows and the columns for node 1 and node 2
        n1 = (element.node_1.id - 1) * 3
        n2 = (element.node_2.id - 1) * 3
        system.system_matrix[n1: n1 + 3, n1: n1 + 3] += element_matrix[0:3, :3]
        system.system_matrix[n1: n1 + 3, n2: n2 + 3] += element_matrix[0:3, 3:]

        system.system_matrix[n2: n2 + 3, n1: n1 + 3] += element_matrix[3:6, :3]
        system.system_matrix[n2: n2 + 3, n2: n2 + 3] += element_matrix[3:6, 3:]

    # returns True if symmetrical.
    if validate:
        assert np.allclose((system.system_matrix.transpose()), system.system_matrix)


def set_displacement_vector(system, nodes_list):
    """
    :param nodes_list: list containing tuples with
    1.the node
    2. the d.o.f. that is set
    :return: Vector with the displacements of the nodes (If displacement is not known, the value is set
    to NaN)
    """
    if system.system_displacement_vector is None:
        system.system_displacement_vector = np.ones(len(system._vertices) * 3) * np.NaN

    for i in nodes_list:
        index = (i[0] - 1) * 3 + i[1] - 1

        try:
            system.system_displacement_vector[index] = 0
        except IndexError as e:
            raise IndexError(e, 'This often occurs if you set supports before the all the elements are modelled. '
                                'First finish the model.')
    return system.system_displacement_vector


def process_conditions(system):
    indexes = []
    # remove the unsolvable values from the matrix and vectors
    for i in range(system.shape_system_matrix):
        if system.system_displacement_vector[i] == 0:
            indexes.append(i)
        else:
            system._remainder_indexes.append(i)

    system.system_displacement_vector = np.delete(system.system_displacement_vector, indexes, 0)
    system.reduced_force_vector = np.delete(system.system_force_vector, indexes, 0)
    system.reduced_system_matrix = np.delete(system.system_matrix, indexes, 0)
    system.reduced_system_matrix = np.delete(system.reduced_system_matrix, indexes, 1)


def process_supports(system):
    for node in system.supports_hinged:
        set_displacement_vector(system, [(node.id, 1), (node.id, 2)])

    for i in range(len(system.supports_roll)):
        set_displacement_vector(system, [(system.supports_roll[i].id, system.supports_roll_direction[i])])

    for node in system.supports_fixed:
        set_displacement_vector(system, [(node.id, 1), (node.id, 2), (node.id, 3)])

    for node, roll in system.supports_spring_x:
        if not roll:
            set_displacement_vector(system, [(node.id, 2)])

    for node, roll in system.supports_spring_z:
        if not roll:
            set_displacement_vector(system, [(node.id, 1)])

    for node, roll in system.supports_spring_y:
        if not roll:
            set_displacement_vector(system, [(node.id, 1), (node.id, 2)])

    for node_id, angle in system.inclined_roll.items():
        for el in system.node_element_map[node_id]:
            if el.node_1.id == node_id:
                el.a1 = el.angle + angle
            elif el.node_2.id == node_id:
                el.a2 = el.angle + angle

            el.compile_kinematic_matrix(el.a1, el.a2, el.l)
            el.compile_stiffness_matrix()





from anastruct.fem.system import SystemElements

ss = SystemElements(EI=5e3, EA=1e5)

ss.add_multiple_elements([[0, 0], [0, 10]], 10)

top_node = ss.nearest_node('y', 10) + 1
ss.add_support_roll(top_node, 1)
ss.add_support_hinged(1)

ss.point_load(top_node, Fy=-1)

if __name__ == "__main__":
    print(ss.solve(geometrical_non_linear=True))
    ss.show_displacement()



import unittest
from anastruct.sectionbase import section_base
from anastruct.fem import system


class SimpleUnitTest(unittest.TestCase):
    def setUp(self):
        section_base.set_database_name('EU')
        section_base.set_unit_system(length='m', mass_unit='kg', force_unit='N')

    def test_HE100B_EU(self):
        # HEB 100 B has mass=20.4kg/m Ax=26.036cm2 Iy=449.5cm4 Iz=167.3cm4

        section_base.set_unit_system(length='m', mass_unit='kg', force_unit='N')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / 26e-4, 1, places=2)
        self.assertAlmostEqual(param['mass'] / 20.4, 1, places=2)
        self.assertAlmostEqual(param['Iy'] / 449.5e-8, 1, places=2)
        self.assertAlmostEqual(param['Iz'] / 167.3e-8, 1, places=2)
        self.assertAlmostEqual(param['swdl'] / 20.4e1, 1, places=2)

        section_base.set_unit_system(length='cm', mass_unit='kg', force_unit='N')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / 26, 1, places=2)
        self.assertAlmostEqual(param['mass'] / 0.204, 1, places=2)
        self.assertAlmostEqual(param['Iy'] / 449.5, 1, places=2)
        self.assertAlmostEqual(param['Iz'] / 167.3, 1, places=2)
        self.assertAlmostEqual(param['swdl'] / 0.204e1, 1, places=2)

        section_base.set_unit_system(length='mm', mass_unit='kg', force_unit='N')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / 26e2, 1, places=2)
        self.assertAlmostEqual(param['mass'] / 0.0204, 1, places=2)
        self.assertAlmostEqual(param['Iy'] / 449.5e4, 1, places=2)
        self.assertAlmostEqual(param['Iz'] / 167.3e4, 1, places=2)
        self.assertAlmostEqual(param['swdl'] / 0.0204e1, 1, places=2)

        section_base.set_unit_system(length='ft', mass_unit='kg', force_unit='N')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / (26e-4 / (0.3048 ** 2)), 1, places=2)
        self.assertAlmostEqual(param['mass'] / (20.4 * 0.3048), 1, places=2)
        self.assertAlmostEqual(param['Iy'] / (449.5e-8 / (0.3048) ** 4), 1, places=2)
        self.assertAlmostEqual(param['Iz'] / (167.3e-8 / (0.3048) ** 4), 1, places=2)
        self.assertAlmostEqual(param['swdl'] / (20.4 * 0.3048 * 10), 1, places=2)

        section_base.set_unit_system(length='inch', mass_unit='kg', force_unit='N')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / (26e-4 / (0.0254 ** 2)), 1, places=2)
        self.assertAlmostEqual(param['mass'] / (20.4 * 0.0254), 1, places=2)
        self.assertAlmostEqual(param['Iy'] / (449.5e-8 / (0.0254) ** 4), 1, places=2)
        self.assertAlmostEqual(param['Iz'] / (167.3e-8 / (0.0254) ** 4), 1, places=2)
        self.assertAlmostEqual(param['swdl'] / (20.4 * 0.0254 * 10), 1, places=2)

        section_base.set_unit_system(length='ft', mass_unit='lb', force_unit='N')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / (26e-4 / (0.3048 ** 2)), 1, places=2)
        self.assertAlmostEqual(param['mass'] / (20.4 * 0.3048 / 0.454), 1, places=2)
        self.assertAlmostEqual(param['Iy'] / (449.5e-8 / (0.3048) ** 4), 1, places=2)
        self.assertAlmostEqual(param['Iz'] / (167.3e-8 / (0.3048) ** 4), 1, places=2)
        self.assertAlmostEqual(param['swdl'] / (20.4 * 0.3048 * 10), 1, places=2)

        section_base.set_unit_system(length='ft', mass_unit='lb', force_unit='lbf')
        param = section_base.get_section_parameters('HE 100 B')
        self.assertAlmostEqual(param['Ax'] / (26e-4 / (0.3048 ** 2)), 1, places=2)
        self.assertAlmostEqual(param['mass'] / (20.4 * 0.3048 / 0.454), 1, places=2)
        self.assertAlmostEqual(param['Iy'] / (449.5e-8 / (0.3048) ** 4), 1, places=2)
        self.assertAlmostEqual(param['Iz'] / (167.3e-8 / (0.3048) ** 4), 1, places=2)
        self.assertAlmostEqual(param['swdl'] / (20.4 * 0.3048 * 10 / 4.448), 1, places=2)

    def test_W12X65_US(self):
        # W12X65 has mass=65[lb/ft] Ax=19.1[inch2] Iy=533[inch4] Iz=174[inch4]
        section_base.set_database_name('US')
        section_base.set_unit_system(length='inch', mass_unit='lb', force_unit='lbf')
        param = section_base.get_section_parameters('W12X65')
        self.assertAlmostEqual(param['Ax'] / 19.1, 1, places=2)
        self.assertAlmostEqual(param['mass'] / (65 / 12), 1, places=2)
        self.assertAlmostEqual(param['Iy'] / 533, 1, places=2)
        self.assertAlmostEqual(param['Iz'] / 174, 1, places=2)
        self.assertAlmostEqual(param['swdl'] / (65 / 12), 1, places=2)

    def test_sectionbase_steel_section_deflection(self):
        ss = system.SystemElements()
        ss.add_multiple_elements([[0, 0], [6, 0]], n=2, steelsection='IPE 300', E=210e9, sw=False, orient='y')
        ss.add_support_hinged(1)
        ss.add_support_hinged(3)
        ss.point_load(2, Fy=-10000)
        ss.solve()
        self.assertAlmostEqual(-0.00256442, ss.get_node_displacements(2)['uy'])

    def test_sectionbase_steel_section_self_weight_reaction(self):
        ss = system.SystemElements()
        ss.add_element([[0, 0], [20, 0]], steelsection='IPE 300', sw=True)
        ss.add_support_hinged(1)
        ss.add_support_hinged(2)
        ss.solve()
        self.assertAlmostEqual(-4224.24, ss.reaction_forces[1].Fz)

    def test_rectangle_section_deflection(self):
        ss = system.SystemElements()
        ss.add_multiple_elements([[0, 0], [2, 0]], n=2, h=0.05, b=0.16, E=210e9, sw=False, orient='y')
        ss.add_support_hinged(1)
        ss.add_support_hinged(3)
        ss.point_load(2, Fy=-1000)
        ss.solve()
        self.assertAlmostEqual(-0.00047619, ss.get_node_displacements(2)['uy'])

    def test_circle_section_deflection(self):
        ss = system.SystemElements()
        ss.add_multiple_elements([[0, 0], [2, 0]], n=2, d=0.07, E=210e9, sw=False)
        ss.add_support_hinged(1)
        ss.add_support_hinged(3)
        ss.point_load(2, Fy=-1000)
        ss.solve()
        self.assertAlmostEqual(-0.00067339, ss.get_node_displacements(2)['uy'])

    def test_rectangle_section_self_weight_reaction(self):
        ss = system.SystemElements()
        ss.add_element([[0, 0], [2, 0]], h=0.1, b=0.1, E=210e9, sw=True, gamma=10000)
        ss.add_support_hinged(1)
        ss.add_support_hinged(2)
        ss.solve()
        self.assertAlmostEqual(-100, ss.reaction_forces[1].Fz)

    def test_circle_section_self_weight_reaction(self):
        ss = system.SystemElements()
        ss.add_element([[0, 0], [2, 0]], d=0.2, E=210e9, sw=True, gamma=10000)
        ss.add_support_hinged(1)
        ss.add_support_hinged(2)
        ss.solve()
        self.assertAlmostEqual(-314.15926535, ss.reaction_forces[1].Fz)

    def test_show_available_sections(self):
        sections = section_base.available_sections
        self.assertEqual(sections[0], 'CAE 100x10')
        self.assertEqual(len(sections), 2145)

    def test_available_units(self):
        self.assertEqual(section_base.available_units['length'][0], 'm')


if __name__ == "__main__":
    unittest.main()



from anastruct.fem.system import SystemElements
# tests parallel direction of the q-load and dead load.

ss = SystemElements()
a = 1
ss.add_element([3, 0], g=a)
ss.add_element([[3, 0], [6, 3]], g=a)
ss.add_element([12, 3], g=a)
ss.add_element([15, 0], g=a)

ss.add_support_hinged(1)
ss.add_support_hinged(5)

if __name__ == "__main__":
    ss.solve()
    print([a[1] for a in ss.get_node_results_system()])
    ss.show_results()


import math
import numpy as np


class Vertex:
    """
    Utility point in 2D.
    """
    def __init__(self, x, y=None):
        """
        :param x: Can be any of int, float, coordinate list, or other vertex.
        :param y: (int, flt)
        """
        if isinstance(x, (tuple, list)):
            self.coordinates = np.array([x[0], x[1]], dtype=np.float32)
        elif isinstance(x, np.ndarray):
            self.coordinates = np.array(x, dtype=np.float32)
        elif type(x) is Vertex:
            self.coordinates = np.array(x.coordinates, dtype=np.float32)
        else:
            self.coordinates = np.array([x, y], dtype=np.float32)

    @property
    def x(self):
        """

        :return:
        """
        return self.coordinates[0]

    @property
    def y(self):
        return self.coordinates[1]

    @property
    def z(self):
        return self.coordinates[1] * -1

    def modulus(self):
        return np.sqrt(np.sum(self.coordinates**2))

    def unit(self):
        return 1 / self.modulus() * self

    def displace_polar(self, alpha, radius, inverse_z_axis=False):
        if inverse_z_axis:
            self.coordinates[0] += math.cos(alpha) * radius
            self.coordinates[1] -= math.sin(alpha) * radius
        else:
            self.coordinates[0] += math.cos(alpha) * radius
            self.coordinates[1] += math.sin(alpha) * radius

    def __add__(self, other):
        if isinstance(other, (tuple, list)):
            other = np.asarray(other)
        if isinstance(other, Vertex):
            other = other.coordinates

        coordinates = self.coordinates + other
        return Vertex(coordinates)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, (tuple, list)):
            other = np.asarray(other)
        if isinstance(other, Vertex):
            other = other.coordinates

        coordinates = self.coordinates - other
        return Vertex(coordinates)

    def __rsub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        if isinstance(other, (tuple, list)):
            other = np.asarray(other)
        if isinstance(other, Vertex):
            other = other.coordinates

        coordinates = self.coordinates * other
        return Vertex(coordinates)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, (tuple, list)):
            other = np.asarray(other)
        if isinstance(other, Vertex):
            other = other.coordinates

        coordinates = self.coordinates / other
        return Vertex(coordinates)

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __str__(self):
        return "Vertex({}, {})".format(self.x, self.y)


def vertex_range(v1, v2, n):
    dv = v2 - v1
    return [v1 + dv * i / n for i in range(n + 1)]





from anastruct.fem.system import SystemElements

"""
Test the primary force vector when applying a q_load at a hinged element.
"""

ss = SystemElements()
ss.add_element([[0, 0], [7, 0]], spring={2: 0})
ss.add_element([7.1, 0])
ss.add_support_fixed([1, 3])
ss.q_load(-10, 1)
ss.solve()
ss.show_reaction_force()
ss.show_bending_moment()




from math import sin, cos
from anastruct.basic import FEMException
import numpy as np
from functools import lru_cache
import copy

try:
    from anastruct.fem.cython.celements import det_shear, det_moment
except ImportError:
    from anastruct.fem.cython.elements import det_shear, det_moment

"""
The matrices underneath are for slender beams, where the most deformation occurs due to bending.
Shear deformation is not taken into account.
"""

CACHE_BOUND = 32000


class Element:
    def __init__(self, id_, EA, EI, l, angle, vertex_1, vertex_2, spring=None):
        """
        :param id_: integer representing the elements ID
        :param EA: Young's modulus * Area
        :param EI: Young's modulus * Moment of Inertia
        :param l: length
        :param angle: angle between element and x-axis
        :param vertex_1: point object
        :param vertex_2: point object
        :param hinge: (integer) 1 or 2. Adds an hinge ad the first or second node.
        :param spring: (dict) Set a spring at node 1 or node 2.
                            {
                              1: { direction: 1
                                   k: 5e3
                                 }
                              2: { direction: 1
                                   k: 5e3
                                 }
                            }
        """
        self.id = id_
        self.type = None
        self.EA = EA
        self.EI = EI
        self.l = l
        self.springs = spring
        self.vertex_1 = vertex_1  # location
        self.vertex_2 = vertex_2  # location
        self.angle = self.a1 = self.a2 = angle
        self.kinematic_matrix = kinematic_matrix(angle, angle, l)
        self.constitutive_matrix = None
        self.stiffness_matrix = None
        self.node_id1 = None  # int
        self.node_id2 = None  # int
        self.node_ids = []
        self.node_map = None
        self.element_displacement_vector = np.empty(6)
        self.element_primary_force_vector = np.zeros(6)  # acting external forces
        self.element_force_vector = None
        self.q_load = 0
        self.q_direction = None
        self.dead_load = 0
        self.N_1 = None
        self.N_2 = None
        self.bending_moment = None
        self.shear_force = None
        self.deflection = None
        self.extension = None
        self.max_deflection = None
        self.nodes_plastic = [False, False]
        self.compile_constitutive_matrix(self.EA, self.EI, l)
        self.compile_stiffness_matrix()
        self.section_name = ''  # needed for element annotation

    @property
    def all_q_load(self):
        if self.q_load is None:
            q = 0
        else:
            if self.q_direction == "x":
                q_factor = -sin(self.angle)
            elif self.q_direction == "y":
                q_factor = cos(self.angle)
            elif self.q_direction == "element" or self.q_direction is None:
                q_factor = 1
            elif self.q_direction is not None:
                raise FEMException('Wrong parameters',
                                   "q-loads direction is not set property. Please choose 'x', 'y', or 'element'")
            q = self.q_load * q_factor

        return q + self.dead_load * cos(self.angle)

    @property
    def node_1(self):
        return self.node_map[self.node_id1]

    @property
    def node_2(self):
        return self.node_map[self.node_id2]

    def determine_force_vector(self):
        self.element_force_vector = np.dot(self.stiffness_matrix, self.element_displacement_vector)
        return self.element_force_vector

    def compile_stiffness_matrix(self):
        self.stiffness_matrix = stiffness_matrix(self.constitutive_matrix, self.kinematic_matrix)

    def compile_kinematic_matrix(self, a1, a2, l):
        self.kinematic_matrix = kinematic_matrix(a1, a2, l)

    def compile_constitutive_matrix(self, EA, EI, l):
        self.constitutive_matrix = constitutive_matrix(EA, EI, l, self.springs)

    def update_stiffness(self, factor, node):
        if node == 1:
            self.constitutive_matrix[1][1] *= factor
            self.constitutive_matrix[1][2] *= factor
            self.constitutive_matrix[2][1] *= factor
        elif node == 2:
            self.constitutive_matrix[1][2] *= factor
            self.constitutive_matrix[2][1] *= factor
            self.constitutive_matrix[2][2] *= factor
        self.compile_stiffness_matrix()

    def compile_geometric_non_linear_stiffness_matrix(self):
        self.compile_stiffness_matrix()
        self.stiffness_matrix += geometric_stiffness_matrix(self.l, self.N_1, self.a1, self.a2)

    def reset(self):
        self.element_displacement_vector = np.zeros(6)
        self.element_primary_force_vector = np.zeros(6)

    def __add__(self, other):
        if self.id != other.id:
            raise FEMException('Wrong element:', 'only elements with the same id can be added.')
        el = copy.deepcopy(self)
        for unit in ['bending_moment', 'shear_force', 'deflection', 'extension', 'N_1', 'N_2']:
            if getattr(el, unit) is None:
                setattr(el, unit, getattr(other, unit))
            else:
                setattr(el, unit, getattr(el, unit) + getattr(other, unit))
        el.max_deflection = other.max_deflection if el.max_deflection is None else \
            max(el.max_deflection, other.max_deflection)

        el.node_map[self.node_id1] = el.node_1 + other.node_1
        el.node_map[self.node_id2] = el.node_2 + other.node_2
        el.node_map[self.node_id1].ux = el.node_1.ux + other.node_1.ux
        el.node_map[self.node_id1].uz = el.node_1.uz + other.node_1.uz
        el.node_map[self.node_id1].phi_y = el.node_1.phi_y + other.node_1.phi_y
        el.node_map[self.node_id2].ux = el.node_2.ux + other.node_2.ux
        el.node_map[self.node_id2].uz = el.node_2.uz + other.node_2.uz
        el.node_map[self.node_id2].phi_y = el.node_2.phi_y + other.node_2.phi_y
        return el


@lru_cache(CACHE_BOUND)
def kinematic_matrix(a1, a2, l):
    """
    Kinematic matrix of an element dependent of the angle ai and the length of the element.

    :param a1: (float) angle with respect to the x axis.
    :param l: (float) Length
    """
    c1 = cos(a1)
    s1 = sin(a1)
    c2 = cos(a2)
    s2 = sin(a2)
    return np.array([[-c1, s1, 0, c2, -s2, 0],
                     [s1 / l, c1 / l, -1, -s2 / l, -c2 / l, 0],
                     [-s1 / l, -c1 / l, 0, s2 / l, c2 / l, 1]])


def constitutive_matrix(EA, EI, l, spring=None):
    """
    :param EA: (float) Young's modules * Area
    :param EI: (float) Young's modules * Moment of Inertia
    :param l: (float) Length
    :param spring: (int) 1 or 2. Apply a hinge on the first of the second node.
    :return: (array)
    """
    matrix = np.array([[EA / l, 0, 0],
                      [0, 4 * EI / l, -2 * EI / l],
                      [0, -2 * EI / l, 4 * EI / l]])

    if spring is not None:
        """
        stiffness matrix K: 
        [[ k, k ]
        [ k, k ]]
        
        flexibility matrix C:
        [[ c, c ]    =   [[ 1/k, 1/k ]
        `[ c, c ]]       [  1/k, 1/k ]]
        
        flexibility matrix c + springs on both sides:
        [[ c11 + 1/k1, c12       ]
        [  c21      , c21 + 1/k2 ]]
        """
        if 1 in spring:
            if spring[1] == 0:  # hinge
                matrix[1][1] = matrix[1][2] = matrix[2][1] = 0
            else:
                matrix[1][1] = 1 / (1 / matrix[1][1] + 1 / spring[1])
                matrix[2][1] = 1 / (1 / matrix[2][1] + 1 / spring[1])
        if 2 in spring:
            if spring[2] == 0:  # hinge
                matrix[1][2] = matrix[2][1] = matrix[2][2] = 0
            else:
                matrix[2][1] = 1 / (1 / matrix[2][1] + 1 / spring[2])
                matrix[1][2] = 1 / (1 / matrix[1][2] + 1 / spring[2])
    return matrix


def stiffness_matrix(var_constitutive_matrix, var_kinematic_matrix):
    kinematic_transposed_times_constitutive = np.dot(var_kinematic_matrix.transpose(), var_constitutive_matrix)
    return np.dot(kinematic_transposed_times_constitutive, var_kinematic_matrix)


def geometric_stiffness_matrix(l, N, a1, a2):
    """

    :param l: (float) Length.
    :param N: (float) Axial force.
    :param a1: (float) angle. (First try 1st order)
    :return: (array)
    """
    c1 = cos(a1)
    s1 = sin(a1)
    c2 = cos(a2)
    s2 = sin(a2)
    # http://people.duke.edu/~hpgavin/cee421/frame-finite-def.pdf
    return N/l * np.array([[6/5*s1**2, -6/5*s1*c1, -l/10*s1, -6/5*s2**2, 6/5*s2*c2, -l/10*s2],
                    [-6/5*s1*c1, 6/5*c1**2, l/10*c1, 6/5*s2*c2, -6/5*c2**2, l/10*c2],
                    [-l/10*s1, l/10*c1, 2*l**2/15, l/10*s2, -l/10*c2, -l**2/30],
                    [-6/5*s1**2, 6/5*s1*c1, l/10*s1, 6/5*s2**2, -6/5*s1*c2, l/10*s2],
                    [6/5*s1*c1, -6/5*c1**2, -l/10*c1, -6/5*s2*c2, 6/5*c2**2, -l/10*c2],
                    [-l/10*s1, l/10*c1, -l**2/30, l/10*s2, -l/10*c2, 2*l**2/15]]) \
                * np.array([1, -1, 1, 1, -1, 1])  # conversion from coordinate system


@lru_cache(CACHE_BOUND)
def det_axial(EA, L, q, x):
    """
    See notebook in: anastruct/fem/background/distributed_ax_force.ipynb

    :param q: (flt)
    :param x: (flt) Location of the axial force
    :param EA: (flt)
    :param L: (flt) Length of the beam
    :return: (flt)
    """
    return EA * (L * q / (2 * EA) - q * x / EA)




import numpy as np
import copy
from anastruct.basic import converge
import logging
from scipy import linalg


def stiffness_adaptation(system, verbosity, max_iter):
    """
    Non linear solver for the nodes by adapting the stiffness of the elements (nodes).

    :param system: (SystemElements)
    :param verbosity: (int)
    :param max_iter: (int)
    :return: (np.array) Vector with displacements.
    """
    system.solve(True, naked=True)
    if verbosity == 0:
        logging.info("Starting stiffness adaptation calculation.")

    # check validity
    assert all([mp > 0 for mpd in system.non_linear_elements.values() for mp in mpd]), \
        "Cannot solve for an mp = 0. If you want a hinge set the spring stiffness equal to 0."

    for c in range(max_iter):
        factors = []

        # update the elements stiffnesses
        for k, v in system.non_linear_elements.items():
            el = system.element_map[k]

            for node_no, mp in v.items():
                if node_no == 1:
                    # Fast Ty
                    m_e = el.element_force_vector[2] + el.element_primary_force_vector[2]
                else:
                    # Fast Ty
                    m_e = el.element_force_vector[5] + el.element_primary_force_vector[5]

                if abs(m_e) > mp:
                    el.nodes_plastic[node_no - 1] = True
                if el.nodes_plastic[node_no - 1]:
                    factor = converge(m_e, mp)
                    factors.append(factor)
                    el.update_stiffness(factor, node_no)

        if not np.allclose(factors, 1, 1e-3):
            system.solve(force_linear=True, naked=True)
        else:
            system.post_processor.node_results_elements()
            system.post_processor.node_results_system()
            system.post_processor.reaction_forces()
            system.post_processor.element_results()
            break

    if c == max_iter - 1:
        logging.warning("Couldn't solve the in the amount of iterations given. max_iter={}".format(max_iter))
    elif verbosity == 0:
        logging.info("Solved in {} iterations".format(c))
    return system.system_displacement_vector


def det_linear_buckling(system):
    """
    Determine linear buckling by solving the generalized eigenvalue problem (k -λkg)x = 0.

    geometrical stiffness matrix at buckling point: Kg = f(N_max)
    1st order forces: N0
    Nmax = λN0
    Kg(Nmax) = λ(Kg(N0) = λKg0

    2nd order analysis is solved by:
    (K + λKg0)U = F

    We are interested in the point that there is nog additional load F and displacement U is possible.
    (K + λKg0)ΔU = ΔF = 0
    (K + λKg0) = 0

    Is the generalized eigenvalue problem:
    (A - λB)x = 0

    :param system: (SystemElements)
    :return: (flt) The factor the loads can be increased until the structure fails due to buckling.
    """
    system.solve()

    # buckling
    k0 = system.reduced_system_matrix * 1.0  # copy

    for el in system.element_map.values():
        el.compile_geometric_non_linear_stiffness_matrix()
        el.reset()

    system.solve()
    kg = system.reduced_system_matrix - k0
    # solve (k -λkg)x = 0

    eigenvalues = np.abs(linalg.eigvals(k0, kg))
    return np.min(eigenvalues)


def geometrically_non_linear(system, verbosity=0, buckling_factor=True, discretize_kwargs=None):
    """

    :param system: (SystemElements)
    :param verbosity: (int)
    :param buckling_factor: (bool)
    :param discretize_kwargs: (dict) Containing the kwargs passed to the discretize function
    :param discretize: (function) discretize function.
    :return: buckling_factor: (flt) The factor the loads can be increased until the structure fails due to buckling.
    """
    # https://www.ethz.ch/content/dam/ethz/special-interest/baug/ibk/structural-mechanics-dam/education/femI/Lecture_2b.pdf
    if verbosity == 0:
        logging.info("Starting geometrical non linear calculation")

    if buckling_factor:
        buckling_system = copy.copy(system)
        if discretize_kwargs is not None:
            buckling_system.discretize(**discretize_kwargs)

        buckling_factor = det_linear_buckling(buckling_system)
    else:
        buckling_factor = None

    system.solve()

    for el in system.element_map.values():
        el.compile_geometric_non_linear_stiffness_matrix()

    system.solve()

    return buckling_factor






import math


def converge(lhs, rhs):
    """
    Determine convergence factor.

    :param lhs: (flt)
    :param rhs: (flt)
    :param div: (flt)
    :return: multiplication factor (flt) ((lhs / rhs) - 1) / div + 1
    """
    lhs = abs(lhs)
    rhs = abs(rhs)
    div = max(lhs, rhs) / min(lhs, rhs) * 2

    return (rhs / lhs - 1) / div + 1


def angle_x_axis(delta_x, delta_z):
    # dot product v_x = [1, 0] ; v = [delta_x, delta_z]
    # dot product = 1 * delta_x + 0 * delta_z -> delta_x
    ai = math.acos(delta_x / math.sqrt(delta_x**2 + delta_z**2))
    if delta_z < 0:
        ai = 2 * math.pi - ai

    return ai



import numpy as np
import math
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from anastruct.basic import find_nearest, rotate_xy
from anastruct.fem.plotter.values import PlottingValues, det_scaling_factor, plot_values_axial_force, \
    plot_values_bending_moment, plot_values_deflection, plot_values_element, plot_values_shear_force

PATCH_SIZE = 0.03


class Plotter(PlottingValues):
    def __init__(self, system, mesh):
        super(Plotter, self).__init__(system, mesh)
        self.system = system
        self.one_fig = None
        self.max_q = 0
        self.max_system_point_load = 0

    def __start_plot(self, figsize):
        plt.close("all")
        self.fig = plt.figure(figsize=figsize)
        self.one_fig = self.fig.add_subplot(111)
        plt.tight_layout()

    def __fixed_support_patch(self, max_val):
        """
        :param max_val: max scale of the plot
        """
        width = height = PATCH_SIZE * max_val
        for node in self.system.supports_fixed:
            support_patch = mpatches.Rectangle((node.vertex.x - width * 0.5, - node.vertex.z - width * 0.5),
                                               width, height, color='r', zorder=9)
            self.one_fig.add_patch(support_patch)

    def __hinged_support_patch(self, max_val):
        """
        :param max_val: max scale of the plot
        """
        radius = PATCH_SIZE * max_val
        for node in self.system.supports_hinged:
            support_patch = mpatches.RegularPolygon((node.vertex.x, node.vertex.y - radius),
                                                    numVertices=3, radius=radius, color='r', zorder=9)
            self.one_fig.add_patch(support_patch)

    def __roll_support_patch(self, max_val):
        """
        :param max_val: max scale of the plot
        """
        radius = PATCH_SIZE * max_val
        count = 0
        for node in self.system.supports_roll:
            direction = self.system.supports_roll_direction[count]
            x1 = np.cos(np.pi) * radius + node.vertex.x + radius
            z1 = np.sin(np.pi) * radius + node.vertex.y
            x2 = np.cos(np.radians(90)) * radius + node.vertex.x + radius
            z2 = np.sin(np.radians(90)) * radius + node.vertex.y
            x3 = np.cos(np.radians(270)) * radius + node.vertex.x + radius
            z3 = np.sin(np.radians(270)) * radius + node.vertex.y

            triangle = np.array([[x1, z1], [x2, z2], [x3, z3]])

            if node.id in self.system.inclined_roll:
                angle = self.system.inclined_roll[node.id]
                triangle = rotate_xy(triangle, angle + np.pi * 0.5)
                support_patch = plt.Polygon(triangle, color='r',
                                            zorder=9)
                self.one_fig.add_patch(support_patch)
                self.one_fig.plot(triangle[1:, 0] - 0.5 * radius * np.sin(angle), triangle[1:, 1] - 0.5 * radius *
                                  np.cos(angle), color='r')

            elif direction == 2:  # horizontal roll
                support_patch = mpatches.RegularPolygon((node.vertex.x, node.vertex.y - radius),
                                                        numVertices=3, radius=radius, color='r', zorder=9)
                self.one_fig.add_patch(support_patch)
                y = -node.vertex.z - 2 * radius
                self.one_fig.plot([node.vertex.x - radius, node.vertex.x + radius], [y, y], color='r')
            elif direction == 1:  # vertical roll
                # translate the support to the node

                support_patch = mpatches.Polygon(triangle, color='r', zorder=9)
                self.one_fig.add_patch(support_patch)

                y = node.vertex.y - radius
                self.one_fig.plot([node.vertex.x + radius * 1.5, node.vertex.x + radius * 1.5], [y, y + 2 * radius],
                                  color='r')
            count += 1

    def __rotating_spring_support_patch(self, max_val):
        """
        :param max_val: max scale of the plot
        """
        radius = PATCH_SIZE * max_val

        for node, _ in self.system.supports_spring_y:
            r = np.arange(0, radius, 0.001)
            theta = 25 * np.pi * r / (0.2 * max_val)
            x = np.cos(theta) * r + node.vertex.x
            y = np.sin(theta) * r - radius + node.vertex.y
            self.one_fig.plot(x, y, color='r', zorder=9)

            # Triangle
            support_patch = mpatches.RegularPolygon((node.vertex.x, node.vertex.y - radius * 3),
                                                    numVertices=3, radius=radius * 0.9, color='r', zorder=9)
            self.one_fig.add_patch(support_patch)

    def __spring_support_patch(self, max_val):
        """
        :param max_val: max scale of the plot
        """
        h = PATCH_SIZE * max_val
        left = -0.5 * h
        right = 0.5 * h
        dh = 0.2 * h

        for node, _ in self.system.supports_spring_z:
            yval = np.arange(0, -9, -1) * dh + node.vertex.y
            xval = np.array([0, 0, left, right, left, right, left, 0, 0]) + node.vertex.x

            self.one_fig.plot(xval, yval, color='r', zorder=10)

            # Triangle
            support_patch = mpatches.RegularPolygon((node.vertex.x, -node.vertex.z - h * 2.6),
                                                    numVertices=3, radius=h * 0.9, color='r', zorder=10)
            self.one_fig.add_patch(support_patch)

        for node, _ in self.system.supports_spring_x:
            xval = np.arange(0, 9, 1) * dh + node.vertex.x
            yval = np.array([0, 0, left, right, left, right, left, 0, 0]) + node.vertex.y

            self.one_fig.plot(xval, yval, color='r', zorder=10)

            # Triangle
            support_patch = mpatches.RegularPolygon((node.vertex.x + h * 1.7, -node.vertex.z - h),
                                                    numVertices=3, radius=h * 0.9, color='r', zorder=10)
            self.one_fig.add_patch(support_patch)

    def __q_load_patch(self, max_val, verbosity):
        """
        :param max_val: max scale of the plot

        xn1;yn1  q-load   xn1;yn1
        -------------------
        |__________________|
        x1;y1  element    x2;y2
        """

        for q_id in self.system.loads_q.keys():
            el = self.system.element_map[q_id]
            if el.q_load > 0:
                direction = 1
            else:
                direction = -1

            h = 0.05 * max_val * abs(el.q_load) / self.max_q
            x1 = el.vertex_1.x
            y1 = el.vertex_1.y
            x2 = el.vertex_2.x
            y2 = el.vertex_2.y

            if el.q_direction == "y":
                ai = 0
            elif el.q_direction == "x":
                ai = 0.5 * np.pi
            else:
                ai = -el.angle

            # - value, because the positive z of the system is opposite of positive y of the plotter
            xn1 = x1 + np.sin(ai) * h * direction
            yn1 = y1 + np.cos(ai) * h * direction
            xn2 = x2 + np.sin(ai) * h * direction
            yn2 = y2 + np.cos(ai) * h * direction
            coordinates = ([x1, xn1, xn2, x2], [y1, yn1, yn2, y2])
            self.one_fig.plot(*coordinates, color='g')
            rec = plt.Polygon(np.vstack(coordinates).T, color='g', alpha=0.3)
            self.one_fig.add_patch(rec)

            if verbosity == 0:
                # arrow
                xa_1 = (x2 - x1) * 0.2 + x1 + np.sin(ai) * 0.8 * h * direction
                ya_1 = (y2 - y1) * 0.2 + y1 + np.cos(ai) * 0.8 * h * direction
                len_x = np.sin(ai - np.pi) * 0.6 * h * direction
                len_y = np.cos(ai - np.pi) * 0.6 * h * direction
                xt = xa_1 + np.sin(-el.angle) * 0.4 * h * direction
                yt = ya_1 + np.cos(-el.angle) * 0.4 * h * direction
                # fc = face color, ec = edge color
                self.one_fig.arrow(xa_1, ya_1, len_x, len_y, head_width=h * 0.25, head_length=0.2 * h, ec='g', fc='g')
                self.one_fig.text(xt, yt, "q=%d" % el.q_load, color='k', fontsize=9, zorder=10)

    @staticmethod
    def __arrow_patch_values(Fx, Fz, node, h):
        """
        :param Fx: (float)
        :param Fz: (float)
        :param node: (Node object)
        :param h: (float) Is a scale variable
        :return: Variables for the matplotlib plotter
        """

        F = (Fx ** 2 + Fz ** 2) ** 0.5
        len_x = Fx / F * h
        len_y = -Fz / F * h
        x = node.vertex.x - len_x * 1.2
        y = node.vertex.y - len_y * 1.2

        return x, y, len_x, len_y, F

    def __point_load_patch(self, max_plot_range, verbosity=0):
        """
        :param max_plot_range: max scale of the plot
        """

        for k in self.system.loads_point:
            Fx, Fz = self.system.loads_point[k]
            F = (Fx ** 2 + Fz ** 2) ** 0.5
            node = self.system.node_map[k]
            h = 0.1 * max_plot_range * F / self.max_system_point_load
            x, y, len_x, len_y, F = self.__arrow_patch_values(Fx, Fz, node, h)

            self.one_fig.arrow(x, y, len_x, len_y, head_width=h * 0.15, head_length=0.2 * h, ec='b', fc='orange',
                               zorder=11)
            if verbosity == 0:
                self.one_fig.text(x, y, "F=%d" % F, color='k', fontsize=9, zorder=10)

    def __moment_load_patch(self, max_val):

        h = 0.2 * max_val
        for k, v in self.system.loads_moment.items():

            node = self.system.node_map[k]
            if v > 0:
                self.one_fig.plot(node.vertex.x, -node.vertex.z, marker=r'$\circlearrowleft$', ms=25,
                                  color='orange')
            else:
                self.one_fig.plot(node.vertex.x, -node.vertex.z, marker=r'$\circlearrowright$', ms=25,
                                  color='orange')
            self.one_fig.text(node.vertex.x + h * 0.2, -node.vertex.z + h * 0.2, "T=%d" % v, color='k',
                              fontsize=9, zorder=10)

    def plot_structure(self, figsize, verbosity, show=False, supports=True, scale=1, offset=(0, 0), gridplot=False,
                       annotations=True):
        """
        :param show: (boolean) if True, plt.figure will plot.
        :param supports: (boolean) if True, supports are plotted.
        :param annotations: (boolean) if True, structure annotations are plotted. It includes section name.
        :return:
        """
        if not gridplot:
            self.__start_plot(figsize)

        x, y = super().structure()

        max_x = np.max(x)
        min_x = np.min(x)
        max_y = np.max(y)
        min_y = np.min(y)

        center_x = (max_x - min_x) / 2 + min_x + -offset[0]
        center_y = (max_y - min_y) / 2 + min_x + -offset[1]

        max_plot_range = max(max_x, max_y)
        ax_range = max_plot_range * scale
        plusxrange = center_x + ax_range
        plusyrange = center_y + ax_range * figsize[1] / figsize[0]
        minxrange = center_x - ax_range
        minyrange = center_y - ax_range * figsize[1] / figsize[0]

        self.one_fig.axis([minxrange, plusxrange, minyrange, plusyrange])

        for el in self.system.element_map.values():
            x_val, y_val = plot_values_element(el)
            self.one_fig.plot(x_val, y_val, color='black', marker='s')

            if verbosity == 0:
                # add node ID to plot
                ax_range = max_plot_range * 0.015
                self.one_fig.text(x_val[0] + ax_range, y_val[0] + ax_range, '%d' % el.node_id1, color='g', fontsize=9,
                                  zorder=10)
                self.one_fig.text(x_val[-1] + ax_range, y_val[-1] + ax_range, '%d' % el.node_id2, color='g', fontsize=9,
                                  zorder=10)

                # add element ID to plot
                factor = 0.02 * self.max_val_structure
                x_val = (x_val[0] + x_val[-1]) / 2 - np.sin(el.angle) * factor
                y_val = (y_val[0] + y_val[-1]) / 2 + np.cos(el.angle) * factor

                self.one_fig.text(x_val, y_val, str(el.id), color='r', fontsize=9, zorder=10)
                
                # add element annotation to plot
                # TODO: check how this holds with multiple structure scales.
                if annotations:
                    x_val += + np.sin(el.angle) * factor * 2.3
                    y_val += - np.cos(el.angle) * factor * 2.3
                    self.one_fig.text(x_val, y_val, el.section_name, color='b', fontsize=9, zorder=10)

        # add supports
        if supports:
            self.__fixed_support_patch(max_plot_range * scale)
            self.__hinged_support_patch(max_plot_range * scale)
            self.__roll_support_patch(max_plot_range * scale)
            self.__rotating_spring_support_patch(max_plot_range * scale)
            self.__spring_support_patch(max_plot_range * scale)

        if verbosity == 0:
            # add_loads
            self.__q_load_patch(max_plot_range, verbosity)
            self.__point_load_patch(max_plot_range, verbosity)
            self.__moment_load_patch(max_plot_range)

        if show:
            self.plot()
        else:
            return self.fig

    def _add_node_values(self, x_val, y_val, value_1, value_2, digits):
        offset = self.max_val_structure * 0.015

        # add value to plot
        self.one_fig.text(x_val[1] - offset, y_val[1] + offset, "%s" % round(value_1, digits),
                          fontsize=9, ha='center', va='center', )
        self.one_fig.text(x_val[-2] - offset, y_val[-2] + offset, "%s" % round(value_2, digits),
                          fontsize=9, ha='center', va='center', )

    def _add_element_values(self, x_val, y_val, value, index, digits=2):
        self.one_fig.text(x_val[index], y_val[index], "%s" % round(value, digits),
                          fontsize=9, ha='center', va='center', )

    def plot_result(self, axis_values, force_1=None, force_2=None, digits=2, node_results=True):
        # plot force
        x_val = axis_values[0]
        y_val = axis_values[1]
        self.one_fig.plot(x_val, y_val, color='b')

        if node_results:
            self._add_node_values(x_val, y_val, force_1, force_2, digits)

    def plot(self):
        plt.show()

    def axial_force(self, factor=None, figsize=None, verbosity=0, scale=1, offset=(0, 0), show=True, gridplot=False):
        self.plot_structure(figsize, 1, scale=scale, offset=offset, gridplot=gridplot)

        node_results = True if verbosity == 0 else False
        if factor is None:
            max_force = max(map(
                lambda el: max(abs(el.N_1), abs(el.N_2)),
                self.system.element_map.values()))
            factor = det_scaling_factor(max_force, self.max_val_structure)

        for el in self.system.element_map.values():
            if math.isclose(el.N_1, 0, rel_tol=1e-5, abs_tol=1e-9):
                continue
            else:
                axis_values = plot_values_axial_force(el, factor)

                self.plot_result(axis_values, el.N_1, el.N_2, node_results=node_results)

                point = (el.vertex_2 - el.vertex_1) / 2 + el.vertex_1
                if el.N_1 < 0:
                    point.displace_polar(alpha=el.angle + 0.5 * np.pi, radius=0.5 * el.N_1 * factor, inverse_z_axis=True)

                    if verbosity == 0:
                        self.one_fig.text(point.x, point.y, "-", ha='center', va='center',
                                          fontsize=20, color='b')
                if el.N_1 > 0:
                    point.displace_polar(alpha=el.angle + 0.5 * np.pi, radius=0.5 * el.N_1 * factor, inverse_z_axis=True)

                    if verbosity == 0:
                        self.one_fig.text(point.x, point.y, "+", ha='center', va='center',
                                          fontsize=14, color='b')

        if show:
            self.plot()
        else:
            return self.fig

    def bending_moment(self, factor=None, figsize=None, verbosity=0, scale=1, offset=(0, 0), show=True, gridplot=False):
        self.plot_structure(figsize, 1, scale=scale, offset=offset, gridplot=gridplot)
        con = len(self.system.element_map[1].bending_moment)
        if factor is None:
            # maximum moment determined by comparing the node's moments and the sagging moments.
            max_moment = max(map(lambda el: max(abs(el.node_1.Ty),
                                                abs((el.node_1.Ty - el.node_2.Ty) * 0.5 - 1 / 8
                                                    * el.all_q_load * el.l**2),) if el.type == 'general' else 0,
                                 self.system.element_map.values()))
            factor = det_scaling_factor(max_moment, self.max_val_structure)

        # determine the axis values
        for el in self.system.element_map.values():
            if math.isclose(el.node_1.Ty, 0, rel_tol=1e-5, abs_tol=1e-9) and \
                    math.isclose(el.node_2.Ty, 0, rel_tol=1e-5, abs_tol=1e-9) and not el.all_q_load:
                # If True there is no bending moment, so no need for plotting.
                continue
            axis_values = plot_values_bending_moment(el, factor, con)
            if verbosity == 0:
                node_results = True
            else:
                node_results = False
            self.plot_result(axis_values, abs(el.node_1.Ty), abs(el.node_2.Ty), node_results=node_results)

            if el.all_q_load:
                m_sag = min(el.bending_moment)
                index = find_nearest(el.bending_moment, m_sag)[1]
                offset = -self.max_val_structure * 0.05

                if verbosity == 0:
                    x = axis_values[0][index] + np.sin(-el.angle) * offset
                    y = axis_values[1][index] + np.cos(-el.angle) * offset
                    self.one_fig.text(x, y, "%s" % round(m_sag, 1),
                                      fontsize=9)
        if show:
            self.plot()
        else:
            return self.fig

    def shear_force(self, factor=None, figsize=None, verbosity=0, scale=1, offset=(0, 0), show=True, gridplot=False):
        self.plot_structure(figsize, 1, scale=scale, offset=offset, gridplot=gridplot)
        if factor is None:
            max_force = max(map(lambda el: np.max(np.abs(el.shear_force)), self.system.element_map.values()))
            factor = det_scaling_factor(max_force, self.max_val_structure)

        for el in self.system.element_map.values():
            if math.isclose(el.node_1.Ty, 0, rel_tol=1e-5, abs_tol=1e-9) and \
                    math.isclose(el.node_2.Ty, 0, rel_tol=1e-5, abs_tol=1e-9) and el.q_load is None:
                # If True there is no bending moment and no shear, thus no shear force, so no need for plotting.
                continue
            axis_values = plot_values_shear_force(el, factor)
            shear_1 = el.shear_force[0]
            shear_2 = el.shear_force[-1]

            if verbosity == 0:
                node_results = True
            else:
                node_results = False

            self.plot_result(axis_values, shear_1, shear_2, node_results=node_results)
        if show:
            self.plot()
        else:
            return self.fig

    def reaction_force(self, figsize, verbosity, scale, offset, show, gridplot=False):
        self.plot_structure(figsize, 1, supports=False, scale=scale, offset=offset, gridplot=gridplot)

        h = 0.2 * self.max_val_structure
        max_force = max(map(lambda node: max(abs(node.Fx), abs(node.Fz)), self.system.reaction_forces.values()))

        for node in self.system.reaction_forces.values():
            if not math.isclose(node.Fx, 0, rel_tol=1e-5, abs_tol=1e-9):
                # x direction
                scale = abs(node.Fx) / max_force * h
                sol = self.__arrow_patch_values(node.Fx, 0, node, scale)
                x = sol[0]
                y = sol[1]
                len_x = sol[2]
                len_y = sol[3]

                self.one_fig.arrow(x, y, len_x, len_y, head_width=h * 0.15, head_length=0.2 * scale, ec='b',
                                   fc='orange',
                                   zorder=11)

                if verbosity == 0:
                    self.one_fig.text(x, y, "R=%s" % round(node.Fx, 2), color='k', fontsize=9, zorder=10)

            if not math.isclose(node.Fz, 0, rel_tol=1e-5, abs_tol=1e-9):
                # z direction
                scale = abs(node.Fz) / max_force * h
                sol = self.__arrow_patch_values(0, node.Fz, node, scale)
                x = sol[0]
                y = sol[1]
                len_x = sol[2]
                len_y = sol[3]

                self.one_fig.arrow(x, y, len_x, len_y, head_width=h * 0.15, head_length=0.2 * scale, ec='b',
                                   fc='orange',
                                   zorder=11)

                if verbosity == 0:
                    self.one_fig.text(x, y, "R=%s" % round(node.Fz, 2), color='k', fontsize=9, zorder=10)

            if not math.isclose(node.Ty, 0, rel_tol=1e-5, abs_tol=1e-9):
                """
                '$...$': render the strings using mathtext
                """
                if node.Ty > 0:
                    self.one_fig.plot(node.vertex.x, -node.vertex.z, marker=r'$\circlearrowleft$', ms=25,
                                      color='orange')
                if node.Ty < 0:
                    self.one_fig.plot(node.vertex.x, -node.vertex.z, marker=r'$\circlearrowright$', ms=25,
                                      color='orange')

                if verbosity == 0:
                    self.one_fig.text(node.vertex.x + h * 0.2, -node.vertex.z + h * 0.2, "T=%s" % round(node.Ty, 2),
                                      color='k', fontsize=9, zorder=10)
        if show:
            self.plot()
        else:
            return self.fig

    def displacements(self, factor=None, figsize=None, verbosity=0, scale=1, offset=(0, 0), show=True, linear=False,
                      gridplot=False):
        self.plot_structure(figsize, 1, scale=scale, offset=offset, gridplot=gridplot)
        if factor is None:
            # needed to determine the scaling factor
            max_displacement = max(map(
                lambda el: max(abs(el.node_1.ux), abs(el.node_1.uz), el.max_deflection) if el.type == 'general' else 0,
                                       self.system.element_map.values()))
            factor = det_scaling_factor(max_displacement, self.max_val_structure)

        for el in self.system.element_map.values():
            axis_values = plot_values_deflection(el, factor, linear)
            self.plot_result(axis_values, node_results=False)

            if el.type == "general":
                # index of the max deflection
                index = np.argmax(np.abs(el.deflection))

                if verbosity == 0:
                    x = (el.node_1.ux + el.node_2.ux) / 2
                    y = (-el.node_1.uz + -el.node_2.uz) / 2

                    if index != 0 or index != el.deflection.size:
                        self._add_element_values(axis_values[0], axis_values[1],
                                                 el.deflection[index] + (x ** 2 + y ** 2) ** 0.5, index, 3)
        if show:
            self.plot()
        else:
            return self.fig

    def results_plot(self, figsize, verbosity, scale, offset, show):
        """
        Aggregate all the plots in one grid plot.

        :param figsize: (tpl)
        :param verbosity: (int)
        :param scale: (flt)
        :param offset: (tpl)
        :param show: (bool)
        :return: Figure or None
        """
        plt.close("all")
        self.fig = plt.figure(figsize=figsize)
        a = 320
        self.one_fig = self.fig.add_subplot(a + 1)
        plt.title("structure")
        self.plot_structure(figsize, verbosity, show=False, scale=scale, offset=offset, gridplot=True)
        self.one_fig = self.fig.add_subplot(a + 2)
        plt.title("bending moment")
        self.bending_moment(None, figsize, verbosity, scale, offset, False, True)
        self.one_fig = self.fig.add_subplot(a + 3)
        plt.title("shear force")
        self.shear_force(None, figsize, verbosity, scale, offset, False, True)
        self.one_fig = self.fig.add_subplot(a + 4)
        plt.title("axial force")
        self.axial_force(None, figsize, verbosity, scale, offset, False, True)
        self.one_fig = self.fig.add_subplot(a + 5)
        plt.title("displacements")
        self.displacements(None, figsize, verbosity, scale, offset, False, False, True)
        self.one_fig = self.fig.add_subplot(a + 6)
        plt.title("reaction force")
        self.reaction_force(figsize, verbosity, scale, offset, False, True)

        if show:
            self.plot()
        else:
            return self.fig



from anastruct.fem.system import SystemElements
from anastruct.material.profile import HEA
from anastruct.material.units import to_kNm2, to_kN


load_factor = 3

E = 210e3
profile = HEA[180]
EA = to_kN(E * profile['A'])
EI = to_kNm2(E * profile['Iy'])
mp = profile["Wy"] * 235 * 1e-6

ss = SystemElements(EA=EA, EI=EI, load_factor=load_factor)
ss.add_element([[0, 0], [0, 4]], mp={2: mp})
ss.add_element([0, 8], mp={1: mp, 2: mp})
ss.add_element([2, 8], mp={1: mp, 2: mp})
ss.add_element([4, 8], mp={1: mp, 2: mp})
ss.add_element([4, 4], mp={1: mp, 2: mp})
ss.add_element([4, 0], mp={1: mp, 2: mp})
ss.add_truss_element([[0, 4], [4, 4]])
ss.add_support_hinged(1)
ss.add_support_fixed(7)

ss.q_load(-20, 3)
ss.q_load(-20, 4)
ss.q_load(-1, 1)
ss.q_load(-1, 2)

if __name__ == "__main__":
    ss.solve()
    ss.show_structure()
    ss.show_displacement()
    ss.show_bending_moment()





from anastruct.fem.system import SystemElements

ss = SystemElements()
ss.add_element([[0, 0], [2, 0]])
ss.add_element([4, 0])
ss.add_element([6, 0])
ss.add_support_fixed([1, 4])

ss.moment_load([2, 3], [20, -20])

if __name__ == "__main__":
    ss.solve()
    ss.show_bending_moment()



import pprint
import copy
from anastruct.basic import args_to_lists


class LoadCase:
    """
    Group different loads in a load case
    """
    def __init__(self, name):
        """
        :param name: (str) Name of the load case
        """
        self.name = name
        self.spec = dict()
        self.c = 0

    def q_load(self, q, element_id, direction="element"):
        """
        Apply a q-load to an element.

        :param element_id: (int/ list) representing the element ID
        :param q: (flt) value of the q-load
        :param direction: (str) "element", "x", "y"
        """
        self.c += 1
        self.spec['q_load-{}'.format(self.c)] = dict(q=q, element_id=element_id, direction=direction)

    def point_load(self, node_id, Fx=0, Fy=0, rotation=0):
        """
        Apply a point load to a node.

        :param node_id: (int/ list) Nodes ID.
        :param Fx: (flt/ list) Force in global x direction.
        :param Fy: (flt/ list) Force in global x direction.
        :param rotation: (flt/ list) Rotate the force clockwise. Rotation is in degrees.
        """
        self.c += 1
        self.spec['point_load-{}'.format(self.c)] = dict(node_id=node_id, Fx=Fx, Fy=Fy, rotation=rotation)

    def moment_load(self, node_id, Ty):
        """
        Apply a moment on a node.

        :param node_id: (int/ list) Nodes ID.
        :param Ty: (flt/ list) Moments acting on the node.
        """
        self.c += 1
        self.spec['moment_load-{}'.format(self.c)] = dict(node_id=node_id, Ty=Ty)

    def dead_load(self, element_id, g):
        """
        Apply a dead load in kN/m on elements.

        :param element_id: (int/ list) representing the element ID
        :param g: (flt/ list) Weight per meter. [kN/m] / [N/m]
        """
        self.c += 1
        self.spec['dead_load-{}'.format(self.c)] = dict(element_id=element_id, g=g)

    def __str__(self):
        return 'Loadcase {}:\n'.format(self.name) + pprint.pformat(self.spec)


class LoadCombination:
    def __init__(self, name):
        self.name = name
        self.spec = dict()

    def add_load_case(self, lc, factor):
        """
        Add a load case to the load combination.

        :param lc: (:class:`anastruct.fem.util.LoadCase`)
        :param factor: (flt) Multiply all the loads in this LoadCase with this factor.
        """
        lc, factor = args_to_lists(lc, factor)
        for i in range(len(lc)):
            self.spec[lc[i].name] = [lc[i], factor[i]]

    def solve(self, system, force_linear=False, verbosity=0, max_iter=200, geometrical_non_linear=False, **kwargs):
        """
        Evaluate the Load Combination.

        :param system: (:class:`anastruct.fem.system.SystemElements`) Structure to apply loads on.
        :param force_linear: (bool) Force a linear calculation. Even when the system has non linear nodes.
        :param verbosity: (int) 0: Log calculation outputs. 1: silence.
        :param max_iter: (int) Maximum allowed iterations.
        :param geometrical_non_linear: (bool) Calculate second order effects and determine the buckling factor.
        :return: (ResultObject)

        Development **kwargs:
            :param naked: (bool) Whether or not to run the solve function without doing post processing.
            :param discretize_kwargs: When doing a geometric non linear analysis you can reduce or increase the number
                                      of elements created that are used for determining the buckling_factor
        """

        results = {}
        for lc, factor in self.spec.values():
            ss = copy.deepcopy(system)

            ss.load_factor = factor
            ss.apply_load_case(lc)
            ss.solve(force_linear, verbosity, max_iter, geometrical_non_linear, **kwargs)
            results[lc.name] = ss

        ss_combination = copy.deepcopy(system)
        for lc_ss in results.values():
            for k in ss_combination.element_map:
                ss_combination.element_map[k] = ss_combination.element_map[k] + lc_ss.element_map[k]

        results['combination'] = ss_combination
        return results




from anastruct.fem.system import SystemElements

ss = SystemElements(mesh=10)
ss.add_element([10, 0])
ss.add_element([20, 0])
ss.q_load(-10, [1, 2])
ss.add_support_hinged([1, 2, 3])

if __name__ == '__main__':
    ss.solve()
    for el in ss.element_map.values():
        print(el.deflection.max())
    ss.show_results()


from anastruct.fem.system import SystemElements

ss = SystemElements()

ss.add_element([0, 10])
ss.add_element([5, 10])
ss.add_element([5, 0])

ss.add_support_hinged(1)
ss.add_support_hinged(4)

ss.point_load(2, Fy=-10, rotation=30)

if __name__ == "__main__":
    ss.show_structure()



from anastruct.fem.system import SystemElements
# tests parallel direction of the q-load and dead load.

ss = SystemElements()

ss.add_element([-2, 2])
ss.add_element([0, 4])
ss.add_element([-2, 6])
ss.add_element([0, 8])

ss.add_support_hinged(1)
ss.add_support_hinged(5)

ss.q_load(1, (1, 2, 3, 4), "x")

if __name__ == "__main__":
    ss.solve()
    print([a[1] for a in ss.get_node_results_system()])
    ss.show_results()


from .element import plot_values_deflection, plot_values_bending_moment, plot_values_axial_force, \
    plot_values_shear_force, plot_values_element
import numpy as np
import math


def det_scaling_factor(max_unit, max_val_structure):
    if math.isclose(max_unit, 0):
        return 0.1
    else:
        return 0.15 * max_val_structure / max_unit


class PlottingValues:
    def __init__(self, system, mesh):
        self.system = system
        self.mesh = max(3, mesh)
        # used for scaling the plotting values.
        self._max_val_structure = None

    @property
    def max_val_structure(self):
        """
        Determine the maximum value of the structures dimensions.
        :return: (flt)
        """
        if self._max_val_structure is None:
            self._max_val_structure = max(
                map(lambda el: max(el.vertex_1.x, el.vertex_2.x, el.vertex_1.y, el.vertex_2.y),
                    self.system.element_map.values()))
        return self._max_val_structure

    def displacements(self, factor, linear):
        if factor is None:
            # needed to determine the scaling factor
            max_displacement = max(map(
                lambda el: max(abs(el.node_1.ux), abs(el.node_1.uz), el.max_deflection) if el.type == 'general' else 0,
                self.system.element_map.values()))
            factor = det_scaling_factor(max_displacement, self.max_val_structure)
        xy = np.hstack([plot_values_deflection(el, factor, linear) for el in self.system.element_map.values()])
        return xy[0, :], xy[1, :]

    def bending_moment(self, factor):
        if factor is None:
            # maximum moment determined by comparing the node's moments and the sagging moments.
            max_moment = max(map(lambda el: max(abs(el.node_1.Ty), abs(
                (el.node_1.Ty - el.node_2.Ty) * 0.5 - 1 / 8 * el.all_q_load * el.l ** 2)),
                                 self.system.element_map.values()))
            factor = det_scaling_factor(max_moment, self.max_val_structure)

        n = len(self.system.element_map[1].bending_moment)
        xy = np.hstack([plot_values_bending_moment(el, factor, n) for el in self.system.element_map.values()])
        return xy[0, :], xy[1, :]

    def axial_force(self, factor):
        if factor is None:
            max_force = max(map(lambda el: max(abs(el.N_1), abs(el.N_2)),
                                self.system.element_map.values()))
            factor = det_scaling_factor(max_force, self.max_val_structure)
        xy = np.hstack([plot_values_axial_force(el, factor) for el in self.system.element_map.values()])
        return xy[0, :], xy[1, :]

    def shear_force(self, factor):
        if factor is None:
            max_force = max(map(lambda el: np.max(np.abs(el.shear_force)), self.system.element_map.values()))
            factor = det_scaling_factor(max_force, self.max_val_structure)
        xy = np.hstack([plot_values_shear_force(el, factor) for el in self.system.element_map.values()])
        return xy[0, :], xy[1, :]

    def structure(self):
        xy = np.hstack([plot_values_element(el) for el in self.system.element_map.values()])
        return xy[0, :], xy[1, :]



from anastruct.fem.system import SystemElements

ss = SystemElements()
ss.add_element(location=[[0, 0], [5, 0]], EA=5e9, EI=8000)
ss.add_element(location=[[5, 0], [5, 5]], EA=5e9, EI=4000)
ss.moment_load(Ty=10, node_id=3)
ss.add_support_hinged(node_id=1)
ss.add_support_hinged(node_id=3)

ss.solve()
ss.show_structure()
ss.show_reaction_force()
ss.show_axial_force()
ss.show_shear_force()
ss.show_bending_moment()
ss.show_displacement()




from anastruct.fem import system as se


def run():
    system = se.SystemElements()
    system.add_element(location=[[3, 4], [0, 0]], EA=5e9, EI=8000)
    system.add_element(location=[[8, 4], [3, 4]], EA=5e9, EI=4000)
    system.show_structure()

    system.q_load(element_id=2, q=-10)

    system.add_support_hinged(node_id=2)
    system.add_support_fixed(node_id=3)

    system.solve()
    system.show_structure()
    system.show_reaction_force()
    system.show_axial_force()
    system.show_shear_force()
    system.show_bending_moment()
    system.show_displacement()


if __name__ == "__main__":
    run()


import math
from anastruct.fem.system import SystemElements

ss = SystemElements(EA=5000)
ss.add_truss_element(location=[[0, 0], [0, 5]])
ss.add_truss_element(location=[[0, 5], [5, 5]])
ss.add_truss_element(location=[[5, 5], [5, 0]])
ss.add_truss_element(location=[[0, 0], [5, 5]], EA=5000 * math.sqrt(2))

ss.add_support_hinged(node_id=1)
ss.add_support_hinged(node_id=4)

ss.point_load(Fx=10, node_id=2)

ss.solve()
ss.show_structure()
ss.show_reaction_force()
ss.show_axial_force()
ss.show_shear_force()
ss.show_displacement(factor=10)




from anastruct.fem.system import SystemElements

"""
Test the primary force vector when applying a q_load at a hinged element.
"""

ss = SystemElements()
ss.add_element([[0, 0], [3.5, 0]])
ss.add_element([7, 0], spring={1: 100})
ss.add_support_fixed([1])
ss.add_support_hinged(3)
ss.point_load(2, Fy=-100)

if __name__ == "__main__":
    ss.solve()
    ss.show_displacement()
    ss.show_reaction_force()
    ss.show_bending_moment()



from anastruct.fem.system import SystemElements

ss = SystemElements(EA=15000, EI=5000)

# Add beams to the system.
ss.add_element(location=[[0, 0], [0, 5]])
ss.add_element(location=[[0, 5], [5, 5]])
ss.add_element(location=[[5, 5], [5, 0]])

# Add a fixed support at node 1.
ss.add_support_fixed(node_id=1)

# Add a rotational spring support at node 4.
ss.add_support_spring(node_id=4, translation=3, k=4000)

# Add loads.
ss.point_load(Fx=30, node_id=2)
ss.q_load(q=-10, element_id=2)

# Solve
ss.solve()

# Get visual results.
ss.show_structure()
ss.show_reaction_force()
ss.show_axial_force()
ss.show_shear_force()
ss.show_bending_moment()
ss.show_displacement()



import numpy as np
import math


def plot_values_deflection(element, factor, linear=False):
    """
    Determine the plotting values for deflection

    :param element: (fem.Element)
    :param factor: (flt) Factor by which to multiply the plotting values perpendicular to the elements axis.
    :param linear: (bool) If True, the bending in between the elements is determined.
    :return: (np.array/ list) x and y values.
    """
    ux1 = element.node_1.ux * factor
    uz1 = -element.node_1.uz * factor
    ux2 = element.node_2.ux * factor
    uz2 = -element.node_2.uz * factor

    x1 = element.vertex_1.x + ux1
    y1 = -element.vertex_1.z + uz1
    x2 = element.vertex_2.x + ux2
    y2 = -element.vertex_2.z + uz2

    if element.type == "general" and not linear:
        n = len(element.deflection)
        x_val = np.linspace(x1, x2, n)
        y_val = np.linspace(y1, y2, n)

        x_val = x_val + element.deflection * math.sin(element.angle) * factor
        y_val = y_val + element.deflection * -math.cos(element.angle) * factor

    else:  # truss element has no bending
        x_val = np.array([x1, x2])
        y_val = np.array([y1, y2])

    return x_val, y_val


def plot_values_bending_moment(element, factor, n):
    """
    :param element: (object) of the Element class
    :param factor: (float) scaling the plot
    :param n: (integer) amount of x-values
    :return:
    """

    # Determine forces for horizontal element.angle = 0
    T_left = element.node_1.Ty
    T_right = -element.node_2.Ty

    sin = math.sin(-element.angle)
    cos = math.cos(-element.angle)

    # apply angle ai
    x1 = element.vertex_1.x + T_left * sin * factor
    y1 = -element.vertex_1.z + T_left * cos * factor
    x2 = element.vertex_2.x + T_right * sin * factor
    y2 = -element.vertex_2.z + T_right * cos * factor

    interpolate = np.linspace(0, 1, n)
    dx = x2 - x1
    dy = y2 - y1

    # determine moment for 0 < x < length of the element
    x_val = x1 + interpolate * dx
    y_val = y1 + interpolate * dy

    if element.q_load or element.dead_load:
        q = element.all_q_load
        x = interpolate * element.l
        q_part = (-0.5 * -q * x**2 + 0.5 * -q * element.l * x)
        x_val += sin * q_part * factor
        y_val += cos * q_part * factor

    x_val = np.append(x_val, element.vertex_2.x)
    y_val = np.append(y_val, -element.vertex_2.z)
    x_val = np.insert(x_val, 0, element.vertex_1.x)
    y_val = np.insert(y_val, 0, -element.vertex_1.z)

    return x_val, y_val


def plot_values_axial_force(element, factor):
    x1 = element.vertex_1.x
    y1 = -element.vertex_1.z
    x2 = element.vertex_2.x
    y2 = -element.vertex_2.z

    N1 = element.N_1
    N2 = element.N_2

    x_1 = x1 + N1 * math.cos(0.5 * math.pi + element.angle) * factor
    y_1 = y1 + N1 * math.sin(0.5 * math.pi + element.angle) * factor
    x_2 = x2 + N2 * math.cos(0.5 * math.pi + element.angle) * factor
    y_2 = y2 + N2 * math.sin(0.5 * math.pi + element.angle) * factor

    x_val = [x1, x_1, x_2, x2]
    y_val = [y1, y_1, y_2, y2]
    return x_val, y_val


def plot_values_shear_force(element, factor):
    x1 = element.vertex_1.x
    y1 = -element.vertex_1.z
    x2 = element.vertex_2.x
    y2 = -element.vertex_2.z

    shear_1 = element.shear_force[0]
    shear_2 = element.shear_force[-1]

    # apply angle ai
    x_1 = x1 + shear_1 * math.sin(-element.angle) * factor
    y_1 = y1 + shear_1 * math.cos(-element.angle) * factor
    x_2 = x2 + shear_2 * math.sin(-element.angle) * factor
    y_2 = y2 + shear_2 * math.cos(-element.angle) * factor

    x_val = np.array([x1, x_1, x_2, x2])
    y_val = np.array([y1, y_1, y_2, y2])
    return x_val, y_val


def plot_values_element(element):
    x_val = [element.vertex_1.x, element.vertex_2.x]
    y_val = [-element.vertex_1.z, -element.vertex_2.z]
    return x_val, y_val



from anastruct.fem.node import Node
from anastruct.vertex import Vertex
from anastruct.basic import FEMException, angle_x_axis


def ensure_single_hinge(system, spring, node_id1, node_id2):
    if spring is not None and 0 in spring:
        """
        Must be one rotational fixed element per node. Thus keep track of the hinges (k == 0).
        """

        for node in range(1, 3):
            if spring[node] == 0:  # node is a hinged node
                if node == 1:
                    node_id = node_id1
                else:
                    node_id = node_id2

                system.node_map[node_id].hinge = True

                if len(system.node_map[node_id].elements) > 0:
                    pass_hinge = not all([el.hinge == node for el in system.node_map[node_id].elements.values()])
                else:
                    pass_hinge = True
                if not pass_hinge:
                    del spring[node]  # too many hinges at that element.


def append_node_id(self, point_1, point_2, node_id1, node_id2):
    if node_id1 not in self.node_map:
        self.node_map[node_id1] = Node(node_id1, vertex=point_1)
    if node_id2 not in self.node_map:
        self.node_map[node_id2] = Node(node_id2, vertex=point_2)


def det_vertices(system, location_list):
    if isinstance(location_list, Vertex):
        point_1 = system._previous_point
        point_2 = Vertex(location_list)
    elif len(location_list) == 1:
        point_1 = system._previous_point
        point_2 = Vertex(location_list[0][0], location_list[0][1])
    elif isinstance(location_list[0], (int, float)):
        point_1 = system._previous_point
        point_2 = Vertex(location_list[0], location_list[1])
    elif isinstance(location_list[0], Vertex):
        point_1 = location_list[0]
        point_2 = location_list[1]
    else:
        point_1 = Vertex(location_list[0][0], location_list[0][1])
        point_2 = Vertex(location_list[1][0], location_list[1][1])
    system._previous_point = point_2

    return point_1, point_2


def det_node_ids(system, point_1, point_2):
    node_ids = []
    for p in (point_1, point_2):
        k = str(p)
        if k in system._vertices:
            node_id = system._vertices[k]
        else:
            node_id = len(system._vertices) + 1
            system._vertices[k] = node_id
        node_ids.append(node_id)
    return node_ids


def support_check(system, node_id):
    if system.node_map[node_id].hinge:
        raise FEMException("Flawed inputs", "You cannot add a support to a hinged node.")


def force_elements_orientation(point_1, point_2, node_id1, node_id2, spring, mp):
    """
    Forces the elements to be in the first and the last quadrant of the unity circle. Meaning the first node is
    always left and the last node is always right. Or the are both on one vertical line.

    The angle of the element will thus always be between -90 till +90 degrees.
    :return: point_1, point_2, node_id1, node_id2, spring, mp, ai
    """
    # determine the angle of the element with the global x-axis
    delta_x = point_2.x - point_1.x
    delta_z = point_2.z - point_1.z  # minus sign to work with an opposite z-axis
    angle = -angle_x_axis(delta_x, delta_z)

    if delta_x < 0:
        # switch points
        point_1, point_2 = point_2, point_1
        node_id1, node_id2 = node_id2, node_id1

        angle = -angle_x_axis(-delta_x, -delta_z)

        if spring is not None:
            assert type(spring) == dict, "The spring parameter should be a dictionary."
            if 1 in spring and 2 in spring:
                spring[1], spring[2] = spring[2], spring[1]
            elif 1 in spring:
                spring[2] = spring.pop(1)
            elif 2 in spring:
                spring[1] = spring.pop(2)

        if mp is not None:
            if 1 in mp and 2 in mp:
                mp[1], mp[2] = mp[2], mp[1]
            elif 1 in mp:
                mp[2] = mp.pop(1)
            elif 2 in mp:
                mp[1] = mp.pop(2)
    return point_1, point_2, node_id1, node_id2, spring, mp, angle



from anastruct.fem.system import SystemElements

"""
Geting section form sectionbase
"""

ss = SystemElements()

ss.add_element([[0, 0], [10, 0]], steelsection='IPE 300', E=210e9, sw=True, orient='z')
ss.add_element([[0, 1], [10, 1]], b=1.0, h=1.0, E=210e9, gamma=20)
ss.add_element([[0, 2], [10, 2]], d=0.6, E=210e9)

if __name__ == '__main__':
    ss.show_structure(verbosity=0, annotations=True)



from anastruct.fem.system import SystemElements

"""
Test dead loads on the structure. TO DO! Distributed axial force
"""

ss = SystemElements()
ss.add_element([[0, 0], [10, 5]], g=1)
ss.add_support_hinged(1)
ss.add_support_roll(2, 2)

if __name__ == '__main__':
    ss.solve()
    ss.show_reaction_force()
    ss.show_axial_force()
    ss.show_bending_moment()


from anastruct.fem.system import SystemElements
"""
Test dead loads on the structure. TO DO! Distributed axial force
"""


ss = SystemElements()
ss.add_element([[0, 6], [5, 4]], g=5)
ss.add_element([7, 3.2], g=5)
ss.add_element([[0, 0], [5, 4]], spring={2: 0})
ss.add_support_hinged([1, 4])

if __name__ == "__main__":
    ss.solve()
    ss.show_axial_force()



import sys
from setuptools import setup

try:
    from Cython.Build import cythonize
    em = cythonize(['anastruct/cython/cbasic.pyx', 'anastruct/fem/cython/celements.pyx'])
except Exception:
    em = []

if sys.version_info[0] == 3 and sys.version_info[1] < 5:
    sys.exit('Sorry, Python < 3.5 is not supported')

setup(
    name='anastruct',
    version='1.1.0a1',
    description='analyse 2D structures.',
    author='Ritchie Vink',
    author_email='ritchie46@gmail.com',
    url='https://ritchievink.com',
    download_url='https://github.com/ritchie46/anaStruct',
    license='GPL-3.0',
    packages=['anastruct', 'anastruct.fem', 'anastruct.fem.system_components', 'anastruct.fem.examples',
              'anastruct.material', 'anastruct.cython', 'anastruct.fem.cython', 'anastruct.fem.plotter',
              'anastruct.fem.util', 'anastruct.sectionbase'],
    package_data={'': ['*.xml']},
    package_dir='',
    install_requires=[
        'matplotlib==3.0.3',
        'numpy==1.15.4',
        'scipy==1.1.0'
    ],
    ext_modules=em

)



class Node:
    def __init__(self, id, Fx=0, Fz=0, Ty=0, ux=0, uz=0, phi_y=0, vertex=None):
        """
        :param id: ID of the node, integer
        :param Fx: Value of Fx
        :param Fz: Value of Fz
        :param Ty: Value of Ty
        :param ux: Value of ux
        :param uz: Value of uz
        :param phi_y: Value of phi
        :param vertex: Point object
        """
        self.id = id
        # forces
        self.Fx = Fx
        self.Fz = Fz
        self.Ty = Ty
        # displacements
        self.ux = ux
        self.uz = uz
        self.phi_y = phi_y
        self.vertex = vertex
        self.hinge = False
        self.elements = {}

    @property
    def Fy(self):
        return -self.Fz

    def __str__(self):
        if self.vertex:
            return "[id = {}, Fx = {}, Fz = {}, Ty = {}, ux = {}, uz = {}, phi_y = {}, x = {}, y = {}]".format(
                self.id, self.Fx, self.Fz, self.Ty, self.ux, self.uz, self.phi_y, self.vertex.x, self.vertex.y)
        else:
            return "[id = {}, Fx = {}, Fz = {}, Ty = {}, ux = {}, uz = {}, phi_y = {}]".format(
                self.id, self.Fx, self.Fz, self.Ty, self.ux, self.uz, self.phi_y)

    def __add__(self, other):
        assert(self.id == other.id), "Cannot add nodes as the ID's don't match. The nodes positions don't match."
        Fx = self.Fx + other.Fx
        Fz = self.Fz + other.Fz
        Ty = self.Ty + other.Ty

        return Node(self.id, Fx, Fz, Ty, self.ux, self.uz, self.phi_y, self.vertex)

    def __sub__(self, other):
        assert (self.id == other.id), "Cannot subtract nodes as the ID's don't match. The nodes positions don't match."
        Fx = self.Fx - other.Fx
        Fz = self.Fz - other.Fz
        Ty = self.Ty - other.Ty

        return Node(self.id, Fx, Fz, Ty, self.ux, self.uz, self.phi_y, self.vertex)

    def reset(self):
        self.Fx = self.Fz = self.Ty = self.ux = self.uz = self.phi_y = 0





import numpy as np
import collections
try:
    from anastruct.cython.cbasic import converge, angle_x_axis
except ImportError:
    from anastruct.cython.basic import converge, angle_x_axis


def find_nearest(array, value):
    """
    :param array: (numpy array object)
    :param value: (float) value searched for
    :return: (tuple) nearest value, index
    """
    # Subtract the value of the value's in the array. Make the values absolute.
    # The lowest value is the nearest.
    index = (np.abs(array-value)).argmin()
    return array[index], index


def integrate_array(y, dx):
    """
    integrate array y * dx
    """
    return np.cumsum(y) * dx


class FEMException(Exception):
    def __init__(self, type_, message):
        self.type = type_
        self.message = message


def args_to_lists(*args):
    arg_lists = []
    for arg in args:
        if isinstance(arg, collections.Iterable) and not isinstance(arg, str):
            arg_lists.append(arg)
        else:
            arg_lists.append([arg])
    lengths = list(map(len, arg_lists))
    n = max(lengths)
    if n == 1:
        return arg_lists

    args = []
    for arg, l in zip(arg_lists, lengths):
        if l == n:
            args.append(arg)
        else:
            args.append([arg[0] for _ in range(n)])
    return args


def rotation_matrix(angle):
    """

    :param angle: (flt) angle in radians
    :return: rotated euclidean xy matrix
    """
    s = np.sin(angle)
    c = np.cos(angle)
    return np.array([[c, -s], [s, c]])


def rotate_xy(a, angle):
    b = np.array(a)
    b[:, 0] -= a[0, 0]
    b[:, 1] -= a[0, 1]
    b = np.dot(b, rotation_matrix(angle))
    b[:, 0] += a[0, 0]
    b[:, 1] += a[0, 1]
    return b



import copy
import math
import numpy as np

from anastruct.fem.node import Node
from anastruct.basic import integrate_array


class SystemLevel:
    def __init__(self, system):
        self.system = system
        # post processor element level
        self.post_el = ElementLevel(self.system)

    def node_results_elements(self):
        """
        Determines the node results on the system level.
        Results placed in SystemElements class: self.node_objects (list).
        """

        for el in self.system.element_map.values():
            # post processor element level
            self.post_el.node_results(el)

    def node_results_system(self):
        for k, v in self.system.node_element_map.items():
            # reset nodes in case of iterative calculation
            self.system.node_map[k].reset()

            if k in self.system.loads_moment:
                self.system.node_map[k].Ty += self.system.loads_moment[k]

            if k in self.system.loads_point:
                Fx, Fz = self.system.loads_point[k]
                self.system.node_map[k].Fx += Fx
                self.system.node_map[k].Fz += Fz

            for i in range(len(v)):
                node = v[i].node_map[k]
                self.system.node_map[k] -= node

                # The displacements are not summarized. Should be assigned only once
                self.system.node_map[k].ux = -node.ux
                self.system.node_map[k].uz = -node.uz
                self.system.node_map[k].phi_y = -node.phi_y

    def reaction_forces(self):
        supports = []
        for node in self.system.supports_fixed:
            supports.append(node.id)
        for node in self.system.supports_hinged:
            supports.append(node.id)
        for node in self.system.supports_roll:
            supports.append(node.id)
        for node, _ in self.system.supports_spring_x:
            supports.append(node.id)
        for node, _ in self.system.supports_spring_z:
            supports.append(node.id)
        for node, _ in self.system.supports_spring_y:
            supports.append(node.id)

        for node_id in supports:
            node = self.system.node_map[node_id]
            node = copy.copy(node)
            self.system.reaction_forces[node_id] = node
            node.Fx *= -1
            node.Fz *= -1
            node.Ty *= -1
            node.ux = None
            node.uz = None
            node.phi_y = None

    def element_results(self):
        """
        Determines the element results for al elements in the system on element level.
        """

        for el in self.system.element_map.values():
            self.post_el.determine_axial_force(el)
            con = self.system.plotter.mesh
            self.post_el.determine_bending_moment(el, con)
            self.post_el.determine_shear_force(el, con)
            self.post_el.determine_displacements(el, con)


class ElementLevel:
    def __init__(self, system):
        self.system = system

    @staticmethod
    def node_results(element):
        """
        Determine node results on the element level.

        """
        # Global coordinates system
        element.node_map[element.node_id1] = Node(
            id=element.node_id1,
            Fx=element.element_force_vector[0] + element.element_primary_force_vector[0],
            Fz=element.element_force_vector[1] + element.element_primary_force_vector[1],
            Ty=element.element_force_vector[2] + element.element_primary_force_vector[2],
            ux=element.element_displacement_vector[0],
            uz=element.element_displacement_vector[1],
            phi_y=element.element_displacement_vector[2],
        )

        element.node_map[element.node_id2] = Node(
            id=element.node_id2,
            Fx=element.element_force_vector[3] + element.element_primary_force_vector[3],
            Fz=element.element_force_vector[4] + element.element_primary_force_vector[4],
            Ty=element.element_force_vector[5] + element.element_primary_force_vector[5],
            ux=element.element_displacement_vector[3],
            uz=element.element_displacement_vector[4],
            phi_y=element.element_displacement_vector[5]
        )

        # Local coordinate system. With inclined supports
        for i in range(1, 3):
            a_n = getattr(element, 'a{}'.format(i))
            if a_n != element.angle:
                node = element.node_map[getattr(element, 'node_id{}'.format(i))]
                angle = a_n - element.angle
                c = np.cos(angle)
                s = np.sin(angle)
                Fx = node.Fx
                Fz = node.Fz
                ux = node.ux
                uz = node.uz
                node.Fz = c * Fz + s * Fx
                node.Fx = -(c * Fx + s * Fz)
                node.ux = c * ux + s * uz
                node.uz = c * uz + s * ux

    @staticmethod
    def determine_axial_force(element):
        N_1 = (math.sin(element.angle) * element.node_1.Fz) + -(math.cos(element.angle) * element.node_1.Fx)
        N_2 = -(math.sin(element.angle) * element.node_2.Fz) + (math.cos(element.angle) * element.node_2.Fx)

        element.N_1 = N_1
        element.N_2 = N_2

    @staticmethod
    def determine_bending_moment(element, con):
        dT = -(element.node_2.Ty + element.node_1.Ty)  # T2 - (-T1)

        iteration_factor = np.linspace(0, 1, con)
        x = iteration_factor * element.l
        m_val = element.node_1.Ty + iteration_factor * dT
        if element.all_q_load:
            q = element.all_q_load
            q_part = (-0.5 * -q * x ** 2 + 0.5 * -q * element.l * x)
            m_val += q_part

        element.bending_moment = m_val

    @staticmethod
    def determine_shear_force(element, con):
        """
        Determines the shear force by differentiating the bending moment.
        :param element: (object) of the Element class
        """
        dV = np.diff(element.bending_moment)
        dx = element.l / (con - 1)
        shear_force = dV / dx

        # Due to differentiation the first and the last values must be corrected.
        correction = shear_force[1] - shear_force[0]
        shear_force = np.insert(shear_force, 0, [shear_force[0] - 0.5 * correction])
        shear_force = np.insert(shear_force, con, [shear_force[-1] + 0.5 * correction])
        element.shear_force = shear_force

    @staticmethod
    def determine_displacements(element, con):
        """
        Determines the displacement by integrating the bending moment.
        :param element: (object) of the Element class

        w = -M''

        This gives you the formula

        w = -aMx +bx + c

        a = already defined by the integral
        b = Scale the slope of the parabola. This is the rotation of the deflection. You can think of this as
            the angle of the deflection beam. By rotating the beam so that the last deflection w = 0 you get the correct
            value for b. w[-1] = 0.
        c = Translate the parabola. Translate it so that w[0] = 0
        """

        if element.type == 'general':
            dx = element.l / (len(element.bending_moment) - 1)

            # Take the average of cumulative integration from both sides. This is due to numerical differences, that
            # would be observable in symmetrical structures.
            phi_neg = -0.5 * (
                        integrate_array(element.bending_moment, dx) + integrate_array(element.bending_moment[::-1],
                                                                                      dx)) / element.EI
            w = integrate_array(phi_neg, dx)

            # Angle between last w and elements axis. The w array will be corrected so that this angle == 0.
            alpha = np.arctan(w[-1] / element.l)

            lx = np.linspace(0, element.l, con)

            w = w - lx * np.sin(alpha)
            element.deflection = -w
            element.max_deflection = np.max(np.abs(w))

        # Extension
        u = 0.5 * (element.N_1 + element.N_2) / element.EA * element.l
        du = u / con
        element.extension = du * (np.arange(con) + 1)



from anastruct.fem.system import SystemElements

ss = SystemElements(plot_backend="plotly")
ss.add_element([[0, 0], [1, 0]])
ss.add_element([2, 0])

ss.add_support_hinged(1)
ss.add_support_spring(3, 2, 3000)
ss.point_load(2, Fy=-1000)
ss.solve()
ss.show_structure()
ss.show_bending_moment()
ss.show_displacement()


import os
import xml.etree.ElementTree as ElementTree
from anastruct.sectionbase import units


class SectionBase:
    """
    Code is extracted from StruPy project.
    """
    available_database_names = ['EU', 'US', 'UK']

    def __init__(self):
        self.current_length_unit = None
        self.current_mass_unit = None
        self.current_force_unit = None
        self.current_database = None
        self.xml_length_unit = None
        self.xml_area_unit = None
        self.xml_weight_unit = None
        self.xml_self_weight_dead_load = None
        self._root = None  # xml root

        self.set_unit_system()

    @property
    def root(self):
        if self._root is None:
            self.set_database_name('EU')
            self.load_data_from_xml()
        return self._root

    @property
    def available_sections(self):
        return list(map(lambda el: el.attrib['sectionname'], self.root.findall("./sectionlist/sectionlist_item")))

    @property
    def available_units(self):
        return {'length': list(units.l_dict.keys()),
                'mass': list(units.m_dict.keys()),
                'force': list(units.f_dict.keys())}

    def set_unit_system(self, length='m', mass_unit='kg', force_unit='N'):
        self.current_length_unit = units.l_dict[length]
        self.current_mass_unit = units.m_dict[mass_unit]
        self.current_force_unit = units.f_dict[force_unit]

    def set_database_name(self, basename):
        if basename == 'EU' or basename == 'UK':
            self.xml_length_unit = units.m
            self.xml_area_unit = units.m
            self.xml_weight_unit = units.kg
            self.xml_self_weight_dead_load = 10. * units.N
            if basename == 'EU':
                self.current_database = 'sectionbase_EuropeanSectionDatabase.xml'
            else:
                self.current_database = 'sectionbase_BritishSectionDatabase.xml'
        elif basename == 'US':
            self.current_database = 'sectionbase_AmericanSectionDatabase.xml'
            self.xml_length_unit = units.ft
            self.xml_area_unit = units.inch
            self.xml_weight_unit = units.lb
            self.xml_self_weight_dead_load = units.lbf

        self.load_data_from_xml()

    def load_data_from_xml(self):
        self._root = ElementTree.parse(os.path.join(os.path.dirname(__file__), 'data', self.current_database)).getroot()

    def get_section_parameters(self, section_name):
        if self.root is None:
            self.set_database_name('EU')
            self.load_data_from_xml()

        element = dict(
            self.root.findall("./sectionlist/sectionlist_item[@sectionname='{}']".format(section_name))[0].items()
        )
        element['swdl'] = element['mass']
        element = self.convert_units(element)
        return element

    def convert_units(self, element):
        lu = self.xml_length_unit / self.current_length_unit  # long unit
        sdu = self.xml_area_unit / self.current_length_unit  # sect dim unit
        wu = self.xml_weight_unit / self.current_mass_unit  # weight unit
        self_weight_dead_load = self.xml_self_weight_dead_load / self.current_force_unit  # self weight dead load unit

        element['mass'] = float(element['mass']) * wu / lu
        element['Ax'] = float(element['Ax']) * sdu ** 2
        element['Iy'] = float(element['Iy']) * sdu ** 4
        element['Iz'] = float(element['Iz']) * sdu ** 4
        element['swdl'] = float(element['swdl']) * self_weight_dead_load / lu
        return element


section_base = SectionBase()



from libc.stdlib cimport abs
from libc.math cimport acos, sqrt, pi, abs

cpdef converge(double lhs, double rhs):
    """
    Determine convergence factor.

    :param lhs: (flt)
    :param rhs: (flt)
    :param div: (flt)
    :return: multiplication factor (flt) ((lhs / rhs) - 1) / div + 1
    """
    lhs = abs(lhs)
    rhs = abs(rhs)
    # cdef double div
    div = max(lhs, rhs) / min(lhs, rhs) * 2.0
    return (abs(rhs) / abs(lhs) - 1.0) / div + 1.0

cpdef angle_x_axis(double delta_x, double delta_z):
    cdef double ai
    # dot product v_x = [1, 0] ; v = [delta_x, delta_z]
    # dot product = 1 * delta_x + 0 * delta_z -> delta_x
    ai = acos(delta_x / sqrt(delta_x**2 + delta_z**2))
    if delta_z < 0:
        ai = 2 * pi - ai

    return ai


from anastruct.fem.system import SystemElements

ss = SystemElements()

ss.add_element([10, 0])
ss.insert_node(1, factor=0.3)
ss.insert_node(2, [5, 5])

if __name__ == "__main__":
    ss.show_structure(values_only=True)
